from oasislmf.pytools.common.data import oasis_float, oasis_int, null_index
from .policy import calc
from .policy_extras import calc as calc_extra
from .common import EXTRA_VALUES, compute_idx_dtype, DEDUCTIBLE, UNDERLIMIT, OVERLIMIT
from .back_allocation import back_alloc_a2, back_alloc_extra_a2, back_alloc_layer, back_alloc_layer_extra

from numba import njit
import numpy as np
import os
import logging
logger = logging.getLogger(__name__)


@njit(cache=True)
def get_base_children(node, children, nodes_array, temp_children_queue):
    """
    fill up temp_children_queue with all the base children of node from index 0 and return the number of base children
    Args:
        node: top node
        children: array of all the children with loss value for each node
        nodes_array: array of information on all nodes
        temp_children_queue: empty array where we write base children.

    Returns:
        number of base children
    """
    len_children = children[node['children']]
    if len_children:
        temp_children_queue[:len_children] = children[node['children'] + 1: node['children'] + len_children + 1]
        temp_children_queue[len_children] = null_index
        queue_end = len_children
        i = 0
        child_i = 0
        while temp_children_queue[i] != null_index:
            parent = nodes_array[temp_children_queue[i]]
            len_children = children[parent['children']]
            if len_children:
                temp_children_queue[queue_end: queue_end + len_children] = children[parent['children'] + 1: parent['children'] + len_children + 1]
                queue_end += len_children
                temp_children_queue[queue_end] = null_index
            else:
                temp_children_queue[child_i] = temp_children_queue[i]
                child_i += 1
            i += 1
    else:
        temp_children_queue[0] = node['node_id']
        child_i = 1
    return child_i


@njit(cache=True)
def first_time_layer(profile_len, base_children_len, temp_children_queue, compute_idx, nodes_array,
                     sidx_indptr, sidx_indexes,
                     loss_indptr, loss_val
                     ):
    """
    first time there is a back allocation with multiple layer, we duplicate loss and extra from layer 1 to the other layers
    """
    for base_child_i in range(base_children_len):
        child = nodes_array[temp_children_queue[base_child_i]]
        child_val_len = sidx_indptr[sidx_indexes[child['node_id']] + 1] - sidx_indptr[sidx_indexes[child['node_id']]]
        child_loss_val_layer_0 = loss_val[loss_indptr[child['loss']]:
                                          loss_indptr[child['loss']] + child_val_len]
        for p in range(1, profile_len):
            loss_indptr[child['loss'] + p] = compute_idx['loss_ptr_i']
            loss_val[compute_idx['loss_ptr_i']: compute_idx['loss_ptr_i'] + child_val_len] = child_loss_val_layer_0
            compute_idx['loss_ptr_i'] += child_val_len


@njit(cache=True)
def first_time_layer_extra(profile_len, base_children_len, temp_children_queue, compute_idx, nodes_array,
                           sidx_indptr, sidx_indexes,
                           loss_indptr, loss_val,
                           extras_indptr, extras_val,
                           ):
    """
    first time there is a back allocation with multiple layer, we duplicate loss and extra from layer 1 to the other layers
    """
    for base_child_i in range(base_children_len):
        child = nodes_array[temp_children_queue[base_child_i]]
        child_val_len = sidx_indptr[sidx_indexes[child['node_id']] + 1] - sidx_indptr[sidx_indexes[child['node_id']]]
        child_loss_val_layer_0 = loss_val[loss_indptr[child['loss']]:
                                          loss_indptr[child['loss']] + child_val_len]
        for p in range(1, profile_len):
            loss_indptr[child['loss'] + p] = compute_idx['loss_ptr_i']
            loss_val[compute_idx['loss_ptr_i']: compute_idx['loss_ptr_i'] + child_val_len] = child_loss_val_layer_0
            compute_idx['loss_ptr_i'] += child_val_len

            extras_indptr[child['extra'] + p] = compute_idx['extras_ptr_i']
            extras_val[compute_idx['extras_ptr_i']: compute_idx['extras_ptr_i'] + child_val_len] = 0
            compute_idx['extras_ptr_i'] += child_val_len


@njit(cache=True, fastmath=True)
def aggregate_children_extras(node, len_children, nodes_array, children, temp_children_queue, compute_idx,
                              temp_node_sidx, sidx_indexes, sidx_indptr, sidx_val, all_sidx,
                              temp_node_loss, loss_indptr, loss_val,
                              temp_node_extras, extras_indptr, extras_val):
    """
    aggregate the loss and extra of the children of the node that is currently computed
    Args:
        node: node that we compute
        len_children: number of children of the node
        nodes_array: array of information on all nodes
        children: array of all the children with loss value for each node
        temp_children_queue: array storing all the base children of the node
        compute_idx: single element named array containing all the pointer needed to tract the computation (compute_idx_dtype)
        temp_node_sidx: dense array to store if sample id has value for this node
        sidx_indexes: index of sidx for nodes
        sidx_indptr: : index to sidx pointer
        sidx_val: sidx values
        all_sidx: list of all sidx in this computation
        temp_node_loss: dense array storing the sum of children loss
        loss_indptr: index to the loss pointer
        loss_val: loss values
        temp_node_extras: dense array storing the sum of children extra
        extras_indptr: index to the extra pointer
        extras_val: extra values
    """
    sidx_created = False
    node_sidx_start = compute_idx['sidx_ptr_i']
    node_sidx_end = 0
    sidx_indexes[node['node_id']] = compute_idx['sidx_i']
    compute_idx['sidx_i'] += 1

    for p in range(node['profile_len']):
        p_temp_node_loss = temp_node_loss[p]
        p_temp_node_extras = temp_node_extras[p]

        for c in range(node['children'] + 1, node['children'] + len_children + 1):
            child = nodes_array[children[c]]
            child_sidx_val = sidx_val[sidx_indptr[sidx_indexes[child['node_id']]]:
                                      sidx_indptr[sidx_indexes[child['node_id']] + 1]]

            if p == 1 and loss_indptr[child['loss'] + p] == loss_indptr[child['loss']]:
                # this is the first time child branch has multiple layer we create views for root children
                base_children_len = get_base_children(child, children, nodes_array, temp_children_queue)
                # print('new layers', child['level_id'], child['agg_id'], node['profile_len'])
                first_time_layer_extra(
                    node['profile_len'], base_children_len, temp_children_queue, compute_idx, nodes_array,
                    sidx_indptr, sidx_indexes,
                    loss_indptr, loss_val,
                    extras_indptr, extras_val,
                )

            child_loss = loss_val[loss_indptr[child['loss'] + p]:
                                  loss_indptr[child['loss'] + p] + child_sidx_val.shape[0]]
            child_extra = extras_val[extras_indptr[child['extra'] + p]:
                                     extras_indptr[child['extra'] + p] + child_sidx_val.shape[0]]
            # print('child', child['level_id'], child['agg_id'], p, loss_indptr[child['loss'] + p], child_loss[0], child['extra'], extras_indptr[child['extra'] + p])
            for indptr in range(child_sidx_val.shape[0]):
                temp_node_sidx[child_sidx_val[indptr]] = True
                p_temp_node_loss[child_sidx_val[indptr]] += child_loss[indptr]
                p_temp_node_extras[child_sidx_val[indptr]] += child_extra[indptr]
        # print('res', p, p_temp_node_loss[-3], p_temp_node_extras[-3])

        loss_indptr[node['loss'] + p] = compute_idx['loss_ptr_i']
        extras_indptr[node['extra'] + p] = compute_idx['extras_ptr_i']
        if sidx_created:
            for node_sidx_cur in range(node_sidx_start, node_sidx_end):
                loss_val[compute_idx['loss_ptr_i']] = p_temp_node_loss[sidx_val[node_sidx_cur]]
                compute_idx['loss_ptr_i'] += 1
                extras_val[compute_idx['extras_ptr_i']] = p_temp_node_extras[sidx_val[node_sidx_cur]]
                compute_idx['extras_ptr_i'] += 1

        else:
            for sidx in all_sidx:
                if temp_node_sidx[sidx]:
                    sidx_val[compute_idx['sidx_ptr_i']] = sidx
                    compute_idx['sidx_ptr_i'] += 1

                    loss_val[compute_idx['loss_ptr_i']] = p_temp_node_loss[sidx]
                    compute_idx['loss_ptr_i'] += 1

                    extras_val[compute_idx['extras_ptr_i']] = p_temp_node_extras[sidx]
                    compute_idx['extras_ptr_i'] += 1

            node_sidx_end = compute_idx['sidx_ptr_i']
            node_val_len = node_sidx_end - node_sidx_start
            sidx_indptr[compute_idx['sidx_i']] = compute_idx['sidx_ptr_i']
            sidx_created = True
    # print('node', node['node_id'], node['agg_id'], loss_indptr[node['loss']], temp_node_loss[:node['profile_len'], -3], temp_node_extras[:node['profile_len'], -3])
    # fill up all layer if necessary
    for layer in range(node['profile_len'], node['layer_len']):
        loss_indptr[node['loss'] + layer] = loss_indptr[node['loss']]
        extras_indptr[node['extra'] + layer] = extras_indptr[node['extra']]

    return node_val_len


@njit(cache=True, fastmath=True)
def aggregate_children(node, len_children, nodes_array, children, temp_children_queue, compute_idx,
                       temp_node_sidx, sidx_indexes, sidx_indptr, sidx_val, all_sidx,
                       temp_node_loss, loss_indptr, loss_val):
    """
    aggregate the loss of the children of the node that is currently computed
    Args:
        node: node that we compute
        len_children: number of children of the node
        nodes_array: array of information on all nodes
        children: array of all the children with loss value for each node
        temp_children_queue: array storing all the base children of the node
        compute_idx: single element named array containing all the pointer needed to tract the computation (compute_idx_dtype)
        temp_node_sidx: dense array to store if sample id has value for this node
        sidx_indexes: index of sidx for nodes
        sidx_indptr: : index to sidx pointer
        sidx_val: sidx values
        all_sidx: list of all sidx in this computation
        temp_node_loss: dense array storing the sum of children loss
        loss_indptr: index to the loss pointer
        loss_val: loss values
    """
    sidx_created = False
    node_sidx_start = compute_idx['sidx_ptr_i']
    node_sidx_end = 0
    sidx_indexes[node['node_id']] = compute_idx['sidx_i']
    compute_idx['sidx_i'] += 1
    for p in range(node['profile_len']):
        p_temp_node_loss = temp_node_loss[p]
        for c in range(node['children'] + 1, node['children'] + len_children + 1):
            child = nodes_array[children[c]]
            child_sidx_val = sidx_val[sidx_indptr[sidx_indexes[child['node_id']]]:
                                      sidx_indptr[sidx_indexes[child['node_id']] + 1]]
            if p == 1 and loss_indptr[child['loss'] + p] == loss_indptr[child['loss']]:
                # this is the first time child branch has multiple layer we create views for root children
                base_children_len = get_base_children(child, children, nodes_array, temp_children_queue)
                first_time_layer(
                    node['profile_len'], base_children_len, temp_children_queue, compute_idx, nodes_array,
                    sidx_indptr, sidx_indexes,
                    loss_indptr, loss_val
                )
            child_loss = loss_val[loss_indptr[child['loss'] + p]:
                                  loss_indptr[child['loss'] + p] + child_sidx_val.shape[0]]

            for indptr in range(child_sidx_val.shape[0]):
                temp_node_sidx[child_sidx_val[indptr]] = True
                p_temp_node_loss[child_sidx_val[indptr]] += child_loss[indptr]

        loss_indptr[node['loss'] + p] = compute_idx['loss_ptr_i']
        if sidx_created:
            for node_sidx_cur in range(node_sidx_start, node_sidx_end):
                loss_val[compute_idx['loss_ptr_i']] = p_temp_node_loss[sidx_val[node_sidx_cur]]
                compute_idx['loss_ptr_i'] += 1

        else:
            for sidx in all_sidx:
                if temp_node_sidx[sidx]:
                    sidx_val[compute_idx['sidx_ptr_i']] = sidx
                    compute_idx['sidx_ptr_i'] += 1
                    temp_node_sidx[sidx] = False

                    loss_val[compute_idx['loss_ptr_i']] = p_temp_node_loss[sidx]
                    compute_idx['loss_ptr_i'] += 1

            node_sidx_end = compute_idx['sidx_ptr_i']
            node_val_len = node_sidx_end - node_sidx_start
            sidx_indptr[compute_idx['sidx_i']] = compute_idx['sidx_ptr_i']
            sidx_created = True

    # fill up all layer if necessary
    for layer in range(node['profile_len'], node['layer_len']):
        loss_indptr[node['loss'] + layer] = loss_indptr[node['loss']]

    return node_val_len


@njit(cache=True)
def set_parent_next_compute(parent_id, child_id, nodes_array, children, computes, compute_idx):
    """
    Set the parent node that needs to be computed at the next level
    Args:
        parent_id: id of the parent
        child_id: id of the child
        nodes_array: array of information on all nodes
        children: array of all the children with loss value for each node
        computes: array of node that need to be computed.
        compute_idx: single element named array containing all the pointer needed to tract the computation (compute_idx_dtype)
    """
    parent = nodes_array[parent_id]
    parent_children_len = children[parent['children']] + 1
    children[parent['children']] = parent_children_len
    children[parent['children'] + parent_children_len] = child_id
    if parent_children_len == 1:  # first time parent is seen
        computes[compute_idx['next_compute_i']] = parent_id
        compute_idx['next_compute_i'] += 1


@njit(cache=True, fastmath=True)
def load_net_value(computes, compute_idx, nodes_array,
                   sidx_indptr, sidx_indexes,
                   loss_indptr, loss_val):
    # we go through the last level node and replace the gross loss by net loss, then reset compute_i to its value
    net_compute_i = 0
    while computes[net_compute_i]:
        node_i, net_compute_i = computes[net_compute_i], net_compute_i + 1
        node = nodes_array[node_i]
        # net loss layer i is initial loss - sum of all layer up to i
        node_val_len = sidx_indptr[sidx_indexes[node['node_id']] + 1] - sidx_indptr[sidx_indexes[node['node_id']]]
        node_ba_val_prev = loss_val[loss_indptr[node['net_loss']]: loss_indptr[node['net_loss']] + node_val_len]
        for layer in range(node['layer_len']):
            node_ba_val_cur = loss_val[loss_indptr[node['loss'] + layer]: loss_indptr[node['loss'] + layer] + node_val_len]
            # print(node['agg_id'], layer, loss_indptr[node['loss'] + layer], node_ba_val_prev, node_ba_val_cur)
            node_ba_val_cur[:] = np.maximum(node_ba_val_prev - node_ba_val_cur, 0)
            node_ba_val_prev = node_ba_val_cur
    compute_idx['level_start_compute_i'] = 0


@njit(cache=True, fastmath=True, error_model="numpy")
def compute_event(compute_info,
                  keep_input_loss,
                  nodes_array,
                  node_parents_array,
                  node_profiles_array,
                  len_array, max_sidx_val, sidx_indexes, sidx_indptr, sidx_val, loss_indptr, loss_val, extras_indptr, extras_val,
                  children,
                  computes,
                  compute_idx,
                  item_parent_i,
                  fm_profile,
                  stepped):
    """
    compute an entire event, result losses are stored inplace in loss_val
    Args:
        compute_info: general information on the computation (financial_structure.compute_info_dtype)
        keep_input_loss: if true compute the net loss instead of the insured loss
        nodes_array: array of information on all nodes
        node_parents_array: array of node to parent node
        node_profiles_array: array of profile for each node
        len_array: length of array needed to store loss and extra as dence array
        max_sidx_val: maximum sidx value
        sidx_indptr: index to sidx pointer
        sidx_indexes: index of sidx for nodes
        sidx_val: sidx values
        loss_indptr: index of loss for nodes
        loss_val: loss values
        extras_indptr: index of extra for nodes
        extras_val: extra values
        children: array of all the children with loss value for each node
        computes: array of node that need to be computed. it is filled up as the computation carries on
        compute_idx: single element named array containing all the pointer needed to tract the computation (compute_idx_dtype)
        item_parent_i: each node may have multiple parent that are selected one after another. This array keeps track of which parent index to select
        fm_profile: array of all the profiles
        stepped: (True or None) determine if this computation contain stepped profile (using None instead of False to allow jit compilation)

    """
    compute_idx['sidx_i'] = compute_idx['next_compute_i']
    compute_idx['sidx_ptr_i'] = compute_idx['loss_ptr_i'] = sidx_indptr[compute_idx['next_compute_i']]
    compute_idx['extras_ptr_i'] = 0
    compute_idx['compute_i'] = 0

    # dense array to store if sample id has value for this node
    temp_node_sidx = np.zeros(len_array, dtype=oasis_int)

    # sparse array to store the loss after profile is applied
    temp_node_loss_sparse = np.zeros(len_array, dtype=oasis_float)

    # sparse array to store the merged loss of all layer
    temp_node_loss_layer_merge = np.zeros(len_array, dtype=oasis_float)

    # sparse array to loss after layer back alloc
    temp_node_loss_layer_ba = np.zeros((compute_info['max_layer'], len_array), dtype=oasis_float)

    # temp_node_loss: dense array storing the sum of children loss, then the loss factor in children back alloc
    temp_node_loss = np.zeros((compute_info['max_layer'], len_array), dtype=np.float64)

    # temp_node_extras: dense array storing the sum of children extra, then the loss factor in children back alloc
    temp_node_extras = np.zeros((compute_info['max_layer'], len_array, 3), dtype=oasis_float)

    # temp_node_extras_layer_merge: sparse array to store the merged extra of all layer
    temp_node_extras_layer_merge = np.zeros((len_array, 3), dtype=oasis_float)

    # temp_node_extras_layer_merge_save: sparse array to keep the value of extra when profile is apply on merged layer
    temp_node_extras_layer_merge_save = np.zeros((len_array, 3), dtype=oasis_float)

    # temp_children_queue: array storing all the base children of the node
    temp_children_queue = np.empty(nodes_array.shape[0], dtype=oasis_int)

    # create all sidx array
    all_sidx = np.empty(max_sidx_val + EXTRA_VALUES, dtype=oasis_int)
    all_sidx[0] = -5
    all_sidx[1] = -3
    all_sidx[2] = -1
    all_sidx[3:] = np.arange(1, max_sidx_val + 1)

    is_allocation_rule_a0 = compute_info['allocation_rule'] == 0
    is_allocation_rule_a1 = compute_info['allocation_rule'] == 1
    is_allocation_rule_a2 = compute_info['allocation_rule'] == 2

    for level in range(compute_info['start_level'], compute_info['max_level'] + 1):  # perform the bottom up loss computation
        # print(level, next_compute_i, compute_i, next_compute_i-compute_i, computes[compute_i:compute_i+2], computes[next_compute_i-2: next_compute_i+1])

        compute_idx['next_compute_i'] += 1  # next_compute_i will stay at 0 wich will stop the while loop and start the next level
        compute_idx['level_start_compute_i'] = compute_idx['compute_i']

        while computes[compute_idx['compute_i']]:
            compute_node = nodes_array[computes[compute_idx['compute_i']]]
            compute_idx['compute_i'] += 1
            # compute loss sums and extra sum
            len_children = children[compute_node['children']]

            # step 1 depending on the number of children, we aggregate or copy the loss of the previous level
            # we fill up correctly temp_node_loss and temp_node_extras
            if len_children:  # if children we aggregate or copy
                if len_children > 1:
                    storage_node = compute_node
                    temp_node_loss.fill(0)
                    if storage_node['extra'] == null_index:
                        node_val_len = aggregate_children(
                            storage_node, len_children, nodes_array, children, temp_children_queue, compute_idx,
                            temp_node_sidx, sidx_indexes, sidx_indptr, sidx_val, all_sidx,
                            temp_node_loss, loss_indptr, loss_val
                        )
                    else:
                        temp_node_extras.fill(0)
                        node_val_len = aggregate_children_extras(
                            storage_node, len_children, nodes_array, children, temp_children_queue, compute_idx,
                            temp_node_sidx, sidx_indexes, sidx_indptr, sidx_val, all_sidx,
                            temp_node_loss, loss_indptr, loss_val,
                            temp_node_extras, extras_indptr, extras_val
                        )
                    node_sidx = sidx_val[compute_idx['sidx_ptr_i'] - node_val_len: compute_idx['sidx_ptr_i']]

                else:  # only 1 child
                    storage_node = nodes_array[children[compute_node['children'] + 1]]
                    # positive sidx are the same as child
                    node_sidx = sidx_val[sidx_indptr[sidx_indexes[storage_node['node_id']]]:sidx_indptr[sidx_indexes[storage_node['node_id']] + 1]]
                    node_val_len = node_sidx.shape[0]

                    if compute_node['profile_len'] > 1 and loss_indptr[storage_node['loss'] + 1] == loss_indptr[storage_node['loss']]:
                        # first time layer, we need to create view for storage_node and copy loss and extra
                        node_loss = loss_val[loss_indptr[storage_node['loss']]:loss_indptr[storage_node['loss']] + node_val_len]

                        for p in range(1, compute_node['profile_len']):
                            loss_indptr[storage_node['loss'] + p] = compute_idx['loss_ptr_i']
                            loss_val[compute_idx['loss_ptr_i']: compute_idx['loss_ptr_i'] + node_val_len] = node_loss
                            compute_idx['loss_ptr_i'] += node_val_len

                        if compute_node['extra'] != null_index:
                            node_extras = extras_val[extras_indptr[storage_node['extra']]:extras_indptr[storage_node['extra']] + node_val_len]
                            for p in range(1, compute_node['profile_len']):
                                extras_indptr[storage_node['extra'] + p] = compute_idx['extras_ptr_i']
                                extras_val[compute_idx['extras_ptr_i']: compute_idx['extras_ptr_i'] + node_val_len] = node_extras
                                compute_idx['extras_ptr_i'] += node_val_len

                        base_children_len = get_base_children(storage_node, children, nodes_array, temp_children_queue)
                        if base_children_len > 1:
                            if compute_node['extra'] != null_index:
                                first_time_layer_extra(
                                    compute_node['profile_len'], base_children_len, temp_children_queue, compute_idx,
                                    nodes_array,
                                    sidx_indptr, sidx_indexes,
                                    loss_indptr, loss_val,
                                    extras_indptr, extras_val,
                                )
                            else:
                                first_time_layer(
                                    compute_node['profile_len'], base_children_len, temp_children_queue, compute_idx,
                                    nodes_array,
                                    sidx_indptr, sidx_indexes,
                                    loss_indptr, loss_val,
                                )

                    if children[storage_node['children']] and compute_node['extra'] != null_index:
                        # child is not base child so back allocation
                        # we need to keep track of extra before profile
                        for p in range(compute_node['profile_len']):
                            if compute_node['cross_layer_profile']:
                                copy_extra = True
                            else:
                                node_profile = node_profiles_array[compute_node['profiles'] + p]
                                copy_extra = node_profile['i_start'] < node_profile['i_end']
                            if copy_extra:
                                node_extras = extras_val[extras_indptr[storage_node['extra'] + p]:
                                                         extras_indptr[storage_node['extra'] + p] + node_val_len]

                                for i in range(node_val_len):
                                    temp_node_extras[p, node_sidx[i]] = node_extras[i]

            else:  # if no children and in compute then layer 1 loss is already set
                # we create space for extras and copy loss from layer 1 to other layer if they exist
                storage_node = compute_node
                node_sidx = sidx_val[sidx_indptr[sidx_indexes[storage_node['node_id']]]:
                                     sidx_indptr[sidx_indexes[storage_node['node_id']] + 1]]
                node_val_len = node_sidx.shape[0]
                node_loss = loss_val[loss_indptr[storage_node['loss']]:
                                     loss_indptr[storage_node['loss']] + node_val_len]

                if compute_node['extra'] != null_index:  # for layer 1 (p=0)
                    extras_indptr[storage_node['extra']] = compute_idx['extras_ptr_i']
                    node_extras = extras_val[compute_idx['extras_ptr_i']: compute_idx['extras_ptr_i'] + node_val_len]
                    node_extras.fill(0)
                    temp_node_extras[0].fill(0)
                    compute_idx['extras_ptr_i'] += node_val_len

                for p in range(1, compute_node['profile_len']):  # if base level already has layers
                    loss_indptr[storage_node['loss'] + p] = compute_idx['loss_ptr_i']
                    loss_val[compute_idx['loss_ptr_i']: compute_idx['loss_ptr_i'] + node_val_len] = node_loss
                    compute_idx['loss_ptr_i'] += node_val_len

                    if compute_node['extra'] != null_index:
                        extras_indptr[storage_node['extra'] + p] = compute_idx['extras_ptr_i']
                        extras_val[compute_idx['extras_ptr_i']: compute_idx['extras_ptr_i'] + node_val_len].fill(0)
                        compute_idx['extras_ptr_i'] += node_val_len
                        temp_node_extras[p].fill(0)

                for layer in range(compute_node['profile_len'], compute_node['layer_len']):
                    # fill up all layer if necessary
                    loss_indptr[storage_node['loss'] + layer] = loss_indptr[storage_node['loss']]
                    if compute_node['extra'] != null_index:
                        extras_indptr[storage_node['extra'] + layer] = extras_indptr[storage_node['extra']]

                if keep_input_loss:
                    loss_indptr[storage_node['net_loss']] = compute_idx['loss_ptr_i']
                    loss_val[compute_idx['loss_ptr_i']: compute_idx['loss_ptr_i'] + node_val_len] = node_loss
                    compute_idx['loss_ptr_i'] += node_val_len

            base_children_len = 0

            # print('level', level, compute_node['agg_id'], loss_indptr[storage_node['loss']], loss_val[loss_indptr[storage_node['loss']]],
            #       node_profiles_array[compute_node['profiles']], compute_node['extra'] != null_index)

            if compute_node['cross_layer_profile']:
                node_profile = node_profiles_array[compute_node['profiles']]
                if node_profile['i_start'] < node_profile['i_end']:
                    if compute_node['extra'] != null_index:
                        temp_node_loss_layer_merge.fill(0)
                        temp_node_extras_layer_merge.fill(0)

                        for l in range(compute_node['layer_len']):
                            temp_node_loss_layer_merge[:node_val_len] += loss_val[
                                loss_indptr[storage_node['loss'] + l]:
                                loss_indptr[storage_node['loss'] + l] + node_val_len
                            ]
                            temp_node_extras_layer_merge[:node_val_len] += extras_val[
                                extras_indptr[storage_node['extra'] + l]:
                                extras_indptr[storage_node['extra'] + l] + node_val_len
                            ]
                        loss_in = temp_node_loss_layer_merge[:node_val_len]
                        loss_out = temp_node_loss_sparse[:node_val_len]
                        temp_node_extras_layer_merge_save[:node_val_len] = temp_node_extras_layer_merge[
                            :node_val_len]  # save values as they are overwriten

                        for profile_index in range(node_profile['i_start'], node_profile['i_end']):
                            calc_extra(fm_profile[profile_index],
                                       loss_out,
                                       loss_in,
                                       temp_node_extras_layer_merge[:, DEDUCTIBLE],
                                       temp_node_extras_layer_merge[:, OVERLIMIT],
                                       temp_node_extras_layer_merge[:, UNDERLIMIT],
                                       stepped)
                        # print(level, compute_node['agg_id'], base_children_len, fm_profile[profile_index]['calcrule_id'],
                        #       loss_indptr[storage_node['loss']], loss_in, '=>', loss_out)
                        # print(temp_node_extras_layer_merge_save[node_sidx[0], DEDUCTIBLE], '=>', temp_node_extras_layer_merge[0, DEDUCTIBLE], extras_indptr[storage_node['extra']])
                        # print(temp_node_extras_layer_merge_save[node_sidx[0], OVERLIMIT], '=>', temp_node_extras_layer_merge[0, OVERLIMIT])
                        # print(temp_node_extras_layer_merge_save[node_sidx[0], UNDERLIMIT], '=>', temp_node_extras_layer_merge[0, UNDERLIMIT])
                        back_alloc_layer_extra(compute_node['layer_len'], node_val_len, storage_node['loss'], storage_node['extra'],
                                               loss_in, loss_out, loss_indptr, loss_val,
                                               temp_node_loss_layer_ba,
                                               extras_indptr, extras_val,
                                               temp_node_extras_layer_merge, temp_node_extras_layer_merge_save
                                               )

                    else:
                        temp_node_loss_layer_merge.fill(0)
                        for l in range(compute_node['layer_len']):
                            temp_node_loss_layer_merge[:node_val_len] += loss_val[
                                loss_indptr[storage_node['loss'] + l]:
                                loss_indptr[storage_node['loss'] + l] + node_val_len
                            ]
                        loss_in = temp_node_loss_layer_merge[:node_val_len]
                        loss_out = temp_node_loss_sparse[:node_val_len]
                        for profile_index in range(node_profile['i_start'], node_profile['i_end']):
                            calc(fm_profile[profile_index],
                                 loss_out,
                                 loss_in,
                                 stepped)
                        # print(level, compute_node['agg_id'], base_children_len, fm_profile[profile_index]['calcrule_id'],
                        #       loss_indptr[storage_node['loss']], loss_in, '=>', loss_out)
                        back_alloc_layer(compute_node['layer_len'], node_val_len, storage_node['loss'],
                                         loss_in, loss_out, loss_indptr, loss_val, temp_node_loss_layer_ba)

            # we apply the policy term on each layer
            for p in range(compute_node['profile_len']):
                if compute_node['cross_layer_profile']:
                    node_profile = node_profiles_array[compute_node['profiles']]
                else:
                    node_profile = node_profiles_array[compute_node['profiles'] + p]

                if node_profile['i_start'] < node_profile['i_end']:
                    loss_in = loss_val[loss_indptr[storage_node['loss'] + p]:
                                       loss_indptr[storage_node['loss'] + p] + node_val_len]
                    loss_out = temp_node_loss_sparse[:node_val_len]

                    if compute_node['extra'] != null_index:
                        extra = extras_val[extras_indptr[storage_node['extra'] + p]:
                                           extras_indptr[storage_node['extra'] + p] + node_val_len]

                        if compute_node['cross_layer_profile']:
                            loss_out = temp_node_loss_layer_ba[p][:node_val_len]
                        else:
                            for profile_index in range(node_profile['i_start'], node_profile['i_end']):
                                calc_extra(fm_profile[profile_index],
                                           loss_out,
                                           loss_in,
                                           extra[:, DEDUCTIBLE],
                                           extra[:, OVERLIMIT],
                                           extra[:, UNDERLIMIT],
                                           stepped)
                                # print(compute_node['level_id'], 'fm_profile', fm_profile[profile_index])
                                # print(level, compute_node['agg_id'], base_children_len, p, fm_profile[profile_index]['calcrule_id'],
                                #       loss_indptr[storage_node['loss'] + p], loss_in, '=>', loss_out)
                                # print(temp_node_extras[p, node_sidx[0], DEDUCTIBLE], '=>', extra[0, DEDUCTIBLE], extras_indptr[storage_node['extra'] + p])
                                # print(temp_node_extras[p, node_sidx[0], OVERLIMIT], '=>', extra[0, OVERLIMIT])
                                # print(temp_node_extras[p, node_sidx[0], UNDERLIMIT], '=>', extra[0, UNDERLIMIT])
                        if not base_children_len:
                            base_children_len = get_base_children(storage_node, children, nodes_array,
                                                                  temp_children_queue)
                            if is_allocation_rule_a2:
                                ba_children_len = base_children_len
                            else:
                                ba_children_len = 1

                        back_alloc_extra_a2(ba_children_len, temp_children_queue, nodes_array, p,
                                            node_val_len, node_sidx, sidx_indptr, sidx_indexes, sidx_val,
                                            loss_in, loss_out, temp_node_loss, loss_indptr, loss_val,
                                            extra, temp_node_extras, extras_indptr, extras_val)
                    else:
                        if compute_node['cross_layer_profile']:
                            loss_out = temp_node_loss_layer_ba[p][:node_val_len]
                        else:
                            for profile_index in range(node_profile['i_start'], node_profile['i_end']):
                                calc(fm_profile[profile_index],
                                     loss_out,
                                     loss_in,
                                     stepped)
                                # print(compute_node['level_id'], 'fm_profile', fm_profile[profile_index])
                                # print(level, compute_node['agg_id'], base_children_len, p, fm_profile[profile_index]['calcrule_id'],
                                #       loss_indptr[storage_node['loss'] + p], loss_in, '=>', loss_out)
                        if not base_children_len:
                            base_children_len = get_base_children(storage_node, children, nodes_array,
                                                                  temp_children_queue)
                            if is_allocation_rule_a2:
                                ba_children_len = base_children_len
                            else:
                                ba_children_len = 1

                        back_alloc_a2(ba_children_len, temp_children_queue, nodes_array, p,
                                      node_val_len, node_sidx, sidx_indptr, sidx_indexes, sidx_val,
                                      loss_in, loss_out, temp_node_loss, loss_indptr, loss_val)

            if level != compute_info['max_level']:
                if compute_node['parent_len']:  # node has direct parent we add it to the compute
                    parent_id = node_parents_array[compute_node['parent']]
                    set_parent_next_compute(
                        parent_id, storage_node['node_id'],
                        nodes_array, children, computes, compute_idx)
                else:  # node has no direct parent, we look into base node for the next parent to put in the compute queue
                    if not base_children_len:
                        base_children_len = get_base_children(storage_node, children, nodes_array, temp_children_queue)

                    for base_child_i in range(base_children_len):
                        child = nodes_array[temp_children_queue[base_child_i]]
                        parent_id = node_parents_array[child['parent'] + item_parent_i[child['node_id']]]
                        item_parent_i[child['node_id']] += 1
                        set_parent_next_compute(
                            parent_id, child['node_id'],
                            nodes_array, children, computes, compute_idx)
            elif is_allocation_rule_a1:
                if not base_children_len:
                    base_children_len = get_base_children(storage_node, children, nodes_array, temp_children_queue)
                if base_children_len > 1:
                    temp_node_loss.fill(0)
                    for base_child_i in range(base_children_len):
                        child = nodes_array[temp_children_queue[base_child_i]]

                        child_sidx_start = sidx_indptr[sidx_indexes[child['node_id']]]
                        child_sidx_end = sidx_indptr[sidx_indexes[child['node_id']] + 1]
                        child_val_len = child_sidx_end - child_sidx_start

                        child_sidx = sidx_val[child_sidx_start: child_sidx_end]
                        child_loss = loss_val[loss_indptr[child['net_loss']]: loss_indptr[child['net_loss']] + child_val_len]

                        for i in range(child_val_len):
                            temp_node_loss[:, child_sidx[i]] += child_loss[i]

                    for p in range(compute_node['profile_len']):
                        node_loss = loss_val[loss_indptr[storage_node['loss'] + p]:
                                             loss_indptr[storage_node['loss'] + p] + node_val_len]
                        for i in range(node_val_len):
                            if node_loss[i]:
                                temp_node_loss[p, node_sidx[i]] = node_loss[i] / temp_node_loss[p, node_sidx[i]]
                            else:
                                temp_node_loss[p, node_sidx[i]] = 0

                        for base_child_i in range(base_children_len):
                            child = nodes_array[temp_children_queue[base_child_i]]

                            child_sidx_start = sidx_indptr[sidx_indexes[child['node_id']]]
                            child_sidx_end = sidx_indptr[sidx_indexes[child['node_id']] + 1]
                            child_val_len = child_sidx_end - child_sidx_start

                            child_sidx = sidx_val[child_sidx_start: child_sidx_end]
                            child_loss = loss_val[loss_indptr[child['loss'] + p]: loss_indptr[child['loss'] + p] + child_val_len]
                            child_net = loss_val[loss_indptr[child['net_loss'] + p]: loss_indptr[child['net_loss'] + p] + child_val_len]

                            for i in range(child_val_len):
                                child_loss[i] = child_net[i] * temp_node_loss[p, child_sidx[i]]
            elif is_allocation_rule_a0:
                if compute_node['node_id'] != storage_node['node_id']:
                    sidx_indexes[compute_node['node_id']] = sidx_indexes[storage_node['node_id']]
                    for p in range(compute_node['profile_len']):
                        loss_indptr[compute_node['loss'] + p] = loss_indptr[storage_node['loss'] + p]

        compute_idx['compute_i'] += 1

    item_parent_i.fill(1)
    # print(compute_info['max_level'], next_compute_i, compute_i, next_compute_i-compute_i, computes[compute_i:compute_i + 2], computes[next_compute_i - 1: next_compute_i + 1])
    if compute_info['allocation_rule'] != 0:
        compute_idx['level_start_compute_i'] = 0


def init_variable(compute_info, max_sidx_val, temp_dir, low_memory):
    """
    extras, loss contains the same index as sidx
    therefore we can use only sidx_indexes to tract the length of each node values
    """
    max_sidx_count = max_sidx_val + EXTRA_VALUES
    len_array = max_sidx_val + 6

    if low_memory:
        sidx_val = np.memmap(os.path.join(temp_dir, "sidx_val.bin"), mode='w+',
                             shape=(compute_info['node_len'] * max_sidx_count), dtype=oasis_int)
        loss_val = np.memmap(os.path.join(temp_dir, "loss_val.bin"), mode='w+',
                             shape=(compute_info['loss_len'] * max_sidx_count), dtype=oasis_float)
        extras_val = np.memmap(os.path.join(temp_dir, "extras_val.bin"), mode='w+',
                               shape=(compute_info['extra_len'] * max_sidx_count, 3), dtype=oasis_float)
    else:
        sidx_val = np.zeros((compute_info['node_len'] * max_sidx_count), dtype=oasis_int)
        loss_val = np.zeros((compute_info['loss_len'] * max_sidx_count), dtype=oasis_float)
        extras_val = np.zeros((compute_info['extra_len'] * max_sidx_count, 3), dtype=oasis_float)

    sidx_indptr = np.zeros(compute_info['node_len'] + 1, dtype=np.int64)
    loss_indptr = np.zeros(compute_info['loss_len'] + 1, dtype=np.int64)
    extras_indptr = np.zeros(compute_info['extra_len'] + 1, dtype=np.int64)

    sidx_indexes = np.empty(compute_info['node_len'], dtype=oasis_int)
    children = np.zeros(compute_info['children_len'], dtype=np.uint32)
    computes = np.zeros(compute_info['compute_len'], dtype=np.uint32)

    pass_through = np.zeros(compute_info['items_len'] + 1, dtype=oasis_float)
    item_parent_i = np.ones(compute_info['items_len'] + 1, dtype=np.uint32)

    compute_idx = np.empty(1, dtype=compute_idx_dtype)[0]
    compute_idx['next_compute_i'] = 0

    return (max_sidx_val, max_sidx_count, len_array, sidx_indexes, sidx_indptr, sidx_val, loss_indptr, loss_val,
            pass_through, extras_indptr, extras_val, children, computes, item_parent_i, compute_idx)


@njit(cache=True)
def reset_variable(children, compute_idx, computes):
    """
    reset the per event array
    Args:
        children: array of all the children with loss value for each node
        compute_idx: single element named array containing all the pointer needed to tract the computation (compute_idx_dtype)
        computes: array of node to compute
    """
    computes[:compute_idx['next_compute_i']].fill(0)
    children.fill(0)
    compute_idx['next_compute_i'] = 0



import tempfile
import logging
import numpy as np
from contextlib import ExitStack

from .financial_structure import create_financial_structure, load_financial_structure
from .stream_sparse import FMReader, EventWriterSparse, EventWriterOrderedOutputSparse
from .compute_sparse import compute_event as compute_event_sparse
from .compute_sparse import init_variable as init_variable_sparse
from .compute_sparse import reset_variable as reset_variable_sparse
from .compute_sparse import load_net_value
from oasislmf.pytools.utils import redirect_logging
from oasislmf.pytools.common.event_stream import init_streams_in, GUL_STREAM_ID, FM_STREAM_ID, LOSS_STREAM_ID


logger = logging.getLogger(__name__)


def run(create_financial_structure_files, **kwargs):
    if create_financial_structure_files:
        create_financial_structure(kwargs['allocation_rule'], kwargs['static_path'])
    else:
        return run_synchronous(**kwargs)


@redirect_logging(exec_name='fmpy')
def run_synchronous(allocation_rule, files_in, files_out, net_loss, storage_method, **kwargs):
    if allocation_rule == 3:
        allocation_rule = 2
    elif allocation_rule == 0 and net_loss is not None:
        raise NotImplementedError("net loss option is not implemented for alloc rule 0")

    if files_out is not None:
        files_out = files_out[0]

    with ExitStack() as stack:
        streams_in, (stream_source_type, stream_agg_type, max_sidx_val) = init_streams_in(files_in, stack)

        if stream_source_type not in [GUL_STREAM_ID, FM_STREAM_ID, LOSS_STREAM_ID]:
            raise Exception(f'unsupported stream_type {stream_source_type} (most probable cause is that the up stream data are incorrect)')

        if storage_method == "sparse":
            run_synchronous_sparse(max_sidx_val, allocation_rule, streams_in=streams_in, files_out=files_out, net_loss=net_loss, stack=stack,
                                   **kwargs)
        else:
            raise ValueError(f"storage_method {storage_method} is not supported for this version")


def run_synchronous_sparse(max_sidx_val, allocation_rule, static_path, streams_in, files_out, low_memory, net_loss, sort_output, stack, **kwargs):
    compute_info, nodes_array, node_parents_array, node_profiles_array, output_array, fm_profile = load_financial_structure(
        allocation_rule, static_path)

    compute_info = compute_info[0]
    stepped = True if compute_info['stepped'] else None  # https://github.com/numba/numba/issues/4108

    if sort_output:
        event_writer_cls = EventWriterOrderedOutputSparse
    else:
        event_writer_cls = EventWriterSparse

    with tempfile.TemporaryDirectory() as tempdir:
        (max_sidx_val, max_sidx_count, len_array, sidx_indexes, sidx_indptr, sidx_val, loss_indptr, loss_val, pass_through,
         extras_indptr, extras_val, children, computes, item_parent_i, compute_idx) = init_variable_sparse(compute_info, max_sidx_val, tempdir, low_memory)

        if allocation_rule == 0:
            pass_through_out = np.zeros_like(pass_through)
        else:
            pass_through_out = pass_through

        keep_input_loss = False
        if compute_info['allocation_rule'] == 1:
            keep_input_loss = True

        if net_loss is None:  # stream out need to provide gross loss
            gross_writer = stack.enter_context(
                event_writer_cls(
                    files_out, nodes_array, output_array, sidx_indexes,
                    sidx_indptr, sidx_val, loss_indptr, loss_val,
                    pass_through_out, max_sidx_val, computes
                )
            )
            net_writer = False
        elif net_loss == '':  # stream out need to provide net_loss instead of gross loss
            gross_writer = False
            net_writer = stack.enter_context(
                event_writer_cls(
                    files_out, nodes_array, output_array, sidx_indexes,
                    sidx_indptr, sidx_val, loss_indptr, loss_val,
                    pass_through_out, max_sidx_val, computes
                )
            )
            keep_input_loss = True
        elif net_loss == '-':  # double stream out, net_loss written to stdout
            gross_writer = False
            if files_out is not None:  # gross loss written out only if files_out path given
                gross_writer = stack.enter_context(
                    event_writer_cls(
                        files_out, nodes_array, output_array, sidx_indexes,
                        sidx_indptr, sidx_val, loss_indptr, loss_val,
                        pass_through_out, max_sidx_val, computes
                    )
                )
            net_writer = stack.enter_context(
                event_writer_cls(
                    None, nodes_array, output_array, sidx_indexes, sidx_indptr,
                    sidx_val, loss_indptr, loss_val, pass_through_out,
                    max_sidx_val, computes
                )
            )
            keep_input_loss = True
        else:  # double stream out, net_loss is the extra path to write the net to
            gross_writer = stack.enter_context(
                event_writer_cls(
                    files_out, nodes_array, output_array, sidx_indexes,
                    sidx_indptr, sidx_val, loss_indptr, loss_val,
                    pass_through_out, max_sidx_val, computes
                )
            )
            net_writer = stack.enter_context(
                event_writer_cls(
                    net_loss, nodes_array, output_array, sidx_indexes,
                    sidx_indptr, sidx_val, loss_indptr, loss_val,
                    pass_through_out, max_sidx_val, computes
                )
            )
            keep_input_loss = True

        fm_reader = FMReader(nodes_array, sidx_indexes, sidx_indptr, sidx_val,
                             loss_indptr, loss_val, pass_through, len_array, computes, compute_idx)

        for i, event_id in enumerate(fm_reader.read_streams(streams_in)):
            try:
                compute_event_sparse(
                    compute_info,
                    keep_input_loss,
                    nodes_array,
                    node_parents_array,
                    node_profiles_array,
                    len_array, max_sidx_val, sidx_indexes, sidx_indptr, sidx_val, loss_indptr, loss_val, extras_indptr, extras_val,
                    children,
                    computes,
                    compute_idx,
                    item_parent_i,
                    fm_profile,
                    stepped)
                if gross_writer:
                    gross_writer.write(event_id, compute_idx)
                if net_writer:
                    load_net_value(computes, compute_idx, nodes_array, sidx_indptr, sidx_indexes, loss_indptr, loss_val)
                    net_writer.write(event_id, compute_idx)
                reset_variable_sparse(children, compute_idx, computes)
            except Exception:
                node = nodes_array[computes[compute_idx['compute_i']]]
                logger.error(f"event index={i} id={event_id}, "
                             f"at node level_id={node['level_id']} agg_id={node['agg_id']}")
                raise



"""
This file defines the functions that maps the supported perils with the correlation settings. This data is usually
obtained from the model_settings.
"""
from typing import Optional

import pandas as pd


def map_data(data: Optional[dict], logger) -> Optional[pd.DataFrame]:
    """
    Maps data from the model settings to to have Peril ID, peril_correlation_group, and damage_correlation_value.

    Args:
        data: (dict) the data loaded from the model settings

    Returns: (pd.DataFrame) the mapped data
    """
    if data is not None:
        supported_perils = data.get("lookup_settings", {}).get("supported_perils", [])
        correlations_legacy = data.get("correlation_settings", [])
        correlation_settings = data.get("model_settings", {}).get("correlation_settings", correlations_legacy)

        for supported_peril in supported_perils:  # supported_perils is expected to be a list of dict
            supported_peril["peril_correlation_group"] = supported_peril.get("peril_correlation_group", 0)

        supported_perils_df = pd.DataFrame(supported_perils)
        correlation_settings_df = pd.DataFrame(correlation_settings)

        if len(correlation_settings_df) > 0:
            # correlations_settings are defined
            if "damage_correlation_value" not in correlation_settings_df.columns:
                logger.info("Correlation settings: No `damage_correlation_value` found")
                correlation_settings_df["damage_correlation_value"] = 0

            if "hazard_correlation_value" not in correlation_settings_df.columns:
                logger.info("Correlation settings: No `hazard_correlation_value` found")
                correlation_settings_df["hazard_correlation_value"] = 0

        # merge allows duplicates of the "peril_correlation_group" in the supported perils
        # merge does not allow duplicates of the "peril_correlation_group" in the correlation settings
        if len(supported_perils_df) > 0 and len(correlation_settings_df) > 0:
            mapped_data = pd.merge(supported_perils_df, correlation_settings_df, on="peril_correlation_group")
            return mapped_data



"""
Contain all common function and attribute to help read the event stream containing the losses
"""

import selectors
from select import select
import sys

import numpy as np
import numba as nb

from .data import oasis_int, oasis_int_size, oasis_float, oasis_float_size

# streams
PIPE_CAPACITY = 65536  # bytes

# stream source type
CDF_STREAM_ID = 0
GUL_STREAM_ID = 1  # deprecated use LOSS_STREAM_ID
FM_STREAM_ID = 2  # deprecated use LOSS_STREAM_ID
LOSS_STREAM_ID = 2
SUMMARY_STREAM_ID = 3

# stream aggregation type (represent the time of aggregation used
ITEM_STREAM = 1
COVERAGE_STREAM = 2


# special sample id
MEAN_IDX = -1
STD_DEV_IDX = -2
TIV_IDX = -3
CHANCE_OF_LOSS_IDX = NUMBER_OF_AFFECTED_RISK_IDX = -4
MAX_LOSS_IDX = -5


def stream_info_to_bytes(stream_source_type, stream_agg_type):
    """
    From Stream source type and aggregation type produce the stream header
    Args:
        stream_source_type (np.int32):
        stream_agg_type (np.int32):

    Returns:
        return bytes
    """
    return np.array([stream_agg_type], '<i4').tobytes()[:3] + np.int8(stream_source_type).tobytes()


def bytes_to_stream_types(stream_header):
    """
    Read the stream header and return the information on stream type
    Args:
        stream_header: bytes

    Returns:
        (stream source type (np.int32), stream aggregation type (np.int32))
    """
    return np.frombuffer(stream_header[3:], 'i1')[0], np.frombuffer(stream_header[:3] + b'\x00', '<i4')[0]


def read_stream_info(stream_obj):
    """
    from open stream object return the information that characterize the stream (stream_source_type, stream_agg_type, len_sample)
    Args:
        stream_obj: open stream
    Returns:
        (stream_source_type, stream_agg_type, len_sample) as np.int32 triplet
    """
    stream_source_type, stream_agg_type = bytes_to_stream_types(stream_obj.read(4))
    len_sample = np.frombuffer(stream_obj.read(4), dtype=np.int32)[0]
    return stream_source_type, stream_agg_type, len_sample


def get_streams_in(files_in, stack):
    if files_in is None:
        streams_in = [sys.stdin.buffer]
    elif isinstance(files_in, list):
        streams_in = [stack.enter_context(open(file_in, 'rb')) for file_in in files_in]
    else:
        streams_in = [stack.enter_context(open(files_in, 'rb'))]
    return streams_in


def get_and_check_header_in(streams_in):
    streams_info = [read_stream_info(stream_in) for stream_in in streams_in]
    if len(set(streams_info)) > 1:
        raise IOError(f"multiple stream type detected in streams {dict(enumerate(streams_info))}")
    return streams_info[0]


def init_streams_in(files_in, stack):
    """
    if files_in use stdin as stream in
    otherwise open each path in files_in, read the header, check that they are the same, and return the streams and their info
    Args:
        files_in: none or a list of path
        stack: contextlib stack to add the open stream to

    Returns:
        list of open streams and their info
    """
    streams_in = get_streams_in(files_in, stack)
    return streams_in, get_and_check_header_in(streams_in)


@nb.jit(nopython=True, cache=True)
def mv_read(byte_mv, cursor, _dtype, itemsize):
    """
    read a certain dtype from numpy byte view starting at cursor, return the value and the index of the end of the object
    Args:
        byte_mv: numpy byte view
        cursor: index of where the object start
        _dtype: data type of the object
        itemsize: size of the data type

    Returns:
        (object value, end of object index)
    """
    return byte_mv[cursor:cursor + itemsize].view(_dtype)[0], cursor + itemsize


@nb.jit(nopython=True, cache=True)
def mv_write(byte_mv, cursor, _dtype, itemsize, value) -> int:
    """
    load an object into the numpy byte view at index cursor, return the index of the end of the object
    Args:
        byte_mv: numpy byte view
        cursor: index of where the object start
        _dtype: data type of the object
        itemsize: size of the data type
        value: value to write

    Returns:
        end of object index
    """
    byte_mv[cursor:cursor + itemsize].view(_dtype)[0] = value
    return cursor + itemsize


@nb.jit(nopython=True, cache=True)
def mv_write_summary_header(byte_mv, cursor, event_id, summary_id, exposure_value) -> int:
    """
    write a summary header to the numpy byte view at index cursor, return the index of the end of the object
    Args:
        byte_mv: numpy byte view
        cursor: index of where the object start
        event_id: event id
        summary_id: summary id
        exposure_value: exposure value

    Returns:
        end of object index
    """
    # print(event_id, summary_id, exposure_value)
    cursor = mv_write(byte_mv, cursor, oasis_int, oasis_int_size, event_id)
    cursor = mv_write(byte_mv, cursor, oasis_int, oasis_int_size, summary_id)
    cursor = mv_write(byte_mv, cursor, oasis_float, oasis_float_size, exposure_value)
    return cursor


@nb.jit(nopython=True, cache=True)
def mv_write_item_header(byte_mv, cursor, event_id, item_id) -> int:
    """
    write a item header to the numpy byte view at index cursor, return the index of the end of the object
    Args:
        byte_mv: numpy byte view
        cursor: index of where the object start
        event_id: event id
        item_id: item id

    Returns:
        end of object index
    """
    # print(event_id, item_id)
    cursor = mv_write(byte_mv, cursor, oasis_int, oasis_int_size, event_id)
    cursor = mv_write(byte_mv, cursor, oasis_int, oasis_int_size, item_id)
    return cursor


@nb.jit(nopython=True, cache=True)
def mv_write_sidx_loss(byte_mv, cursor, sidx, loss) -> int:
    """
    write sidx and loss to the numpy byte view at index cursor, return the index of the end of the object
    Args:
        byte_mv: numpy byte view
        cursor: index of where the object start
        sidx: sample id
        loss: loss

    Returns:
        end of object index
    """
    # print('    ', sidx, loss)
    cursor = mv_write(byte_mv, cursor, oasis_int, oasis_int_size, sidx)
    cursor = mv_write(byte_mv, cursor, oasis_float, oasis_float_size, loss)
    return cursor


@nb.jit(nopython=True, cache=True)
def mv_write_delimiter(byte_mv, cursor) -> int:
    """
    write the item delimiter (0,0) to the numpy byte view at index cursor, return the index of the end of the object
    Args:
        byte_mv: numpy byte view
        cursor: index of where the object start

    Returns:
        end of delimiter index
    """
    cursor = mv_write(byte_mv, cursor, oasis_int, oasis_int_size, 0)
    cursor = mv_write(byte_mv, cursor, oasis_float, oasis_float_size, 0)
    # print('end', cursor)
    return cursor


class EventReader:
    """
    Abstract class to read event stream

    This class provide a generic interface to read multiple event stream using:
    - selector : handle back pressure, the program is paused and don't use resource if nothing is in the stream buffer
    - memoryview : read a chuck (PIPE_CAPACITY) of data at a time then work on it using a numpy byte view of this buffer

    To use those methods need to be implemented:
    - __init__(self, ...) the constructor with all data structure needed to read and store the event stream
    - read_buffer(self, byte_mv, cursor, valid_buff, event_id, item_id)
        simply point to a local numba.jit function name read_buffer (a template is provided bellow)
        this function should implement the specific logic of where and how to store the event information.

    Those to method may be overwritten
    - item_exit(self):
        specific logic to do when an item is finished (only executed once the stream is finished but no 0,0 closure was present)

    - event_read_log(self):
        what kpi to log when a full event is read

    usage snippet:
        with ExitStack() as stack:
            streams_in, (stream_type, stream_agg_type, len_sample) = init_streams_in(files_in, stack)
            reader = CustomReader(<read relevant attributes>)
            for event_id in reader.read_streams(streams_in):
                <event logic>

    """

    @staticmethod
    def register_streams_in(selector_class, streams_in):
        """
        Data from input process is generally sent by event block, meaning once a stream receive data, the complete event is
        going to be sent in a short amount of time.
        Therefore, we can focus on each stream one by one using their specific selector 'stream_selector'.

        """
        main_selector = selector_class()
        stream_data = []
        for stream_in in streams_in:
            mv = memoryview(bytearray(PIPE_CAPACITY))
            byte_mv = np.frombuffer(buffer=mv, dtype='b')

            stream_selector = selector_class()
            stream_selector.register(stream_in, selectors.EVENT_READ)
            data = {'mv': mv,
                    'byte_mv': byte_mv,
                    'cursor': 0,
                    'valid_buff': 0,
                    'stream_selector': stream_selector
                    }
            stream_data.append(data)
            main_selector.register(stream_in, selectors.EVENT_READ, data)
        return main_selector, stream_data

    def read_streams(self, streams_in):
        """
        read multiple stream input, yield each event id and load relevant value according to the specific read_buffer implemented in subclass
        Args:
            streams_in: streams to read

        Returns:
            event id generator
        """
        try:
            main_selector, stream_data = self.register_streams_in(selectors.DefaultSelector, streams_in)
            self.logger.debug("Streams read with DefaultSelector")
        except PermissionError:  # Fall back option if stream_in contain regular files
            main_selector, stream_data = self.register_streams_in(selectors.SelectSelector, streams_in)
            self.logger.debug("Streams read with SelectSelector")
        try:
            while main_selector.get_map():
                for sKey, _ in main_selector.select():
                    event = self.read_event(sKey.fileobj, main_selector, **sKey.data)

                    if event:
                        event_id, cursor, valid_buff = event
                        sKey.data['cursor'] = cursor
                        sKey.data['valid_buff'] = valid_buff
                        self.event_read_log(event_id)
                        yield event_id

            # Stream is read, we need to check if there is remaining event to be parsed
            for data in stream_data:
                if data['cursor'] < data['valid_buff']:
                    byte_mv = data['byte_mv']
                    cursor = data['cursor']
                    valid_buff = data['valid_buff']
                    yield_event = True
                    while yield_event:
                        cursor, event_id, item_id, yield_event = self.read_buffer(byte_mv, cursor, valid_buff, 0, 0)

                        if event_id:
                            self.item_exit()
                            self.event_read_log(event_id)
                            yield event_id

        finally:
            main_selector.close()

    def read_event(self, stream_in, main_selector, stream_selector, mv, byte_mv, cursor, valid_buff):
        """
        read one event from stream_in
        close and remove the stream from main_selector when all is read
        Args:
            stream_in: stream to read
            main_selector: selector that contain all the streams
            stream_selector:  this stream selector
            mv: buffer memoryview
            byte_mv: numpy byte view of the buffer
            cursor: current cursor of the memory view
            valid_buff: valid data in memory view

        Returns:
            event_id, cursor, valid_buff
        """
        event_id = 0
        item_id = 0

        while True:
            if valid_buff < PIPE_CAPACITY:
                stream_selector.select()
                len_read = stream_in.readinto1(mv[valid_buff:])
                valid_buff += len_read

                if len_read == 0:
                    stream_selector.close()
                    main_selector.unregister(stream_in)
                    if event_id:
                        self.item_exit()
                        return event_id, cursor, valid_buff

                    break
            cursor, event_id, item_id, yield_event = self.read_buffer(byte_mv, cursor, valid_buff, event_id, item_id)

            if yield_event:
                if 2 * cursor > valid_buff:
                    mv[:valid_buff - cursor] = mv[cursor: valid_buff]
                    valid_buff -= cursor
                    cursor = 0
                return event_id, cursor, valid_buff
            else:
                mv[:valid_buff - cursor] = mv[cursor: valid_buff]
                valid_buff -= cursor
                cursor = 0

    def read_buffer(self, byte_mv, cursor, valid_buff, event_id, item_id):
        raise NotImplementedError

    def item_exit(self):
        pass

    def event_read_log(self, event_id):
        pass


# read_buffer template,
# add you input argument and implement what is needed in the three steps ('do new item setup', 'do loss read', 'do item exit')
#
# @nb.jit(nopython=True, cache=True)
# def read_buffer(byte_mv, cursor, valid_buff, event_id, item_id, [array necessary to load and store the event data]):
#
#     last_event_id = event_id
#     while True:
#         if item_id:
#             if valid_buff -  cursor < (oasis_int_size + oasis_float_size):
#                 break
#             sidx, cursor = mv_read(byte_mv, cursor, oasis_int, oasis_int_size)
#             if sidx:
#                 loss, cursor = mv_read(byte_mv, cursor, oasis_float, oasis_float_size)
#                 loss = 0 if np.isnan(loss) else loss
#
#                 ###### do loss read ######
#                 ##########
#
#             else:
#                 ##### do item exit ####
#                 ##########
#                 cursor += oasis_float_size
#                 item_id = 0
#         else:
#             if valid_buff -  cursor < 2 * oasis_int_size:
#                 break
#             event_id, cursor = mv_read(byte_mv, cursor, oasis_int, oasis_int_size)
#             if event_id != last_event_id:
#                 if last_event_id: # we have a new event we return the one we just finished
#                     return cursor - oasis_int_size, last_event_id, 0, 1
#                 else: # first pass we store the event we are reading
#                     last_event_id = event_id
#             item_id, cursor = mv_read(byte_mv, cursor, oasis_int, oasis_int_size)
#
#             ##### do new item setup #####
#             ##########
#     return cursor, event_id, item_id, 0


def write_mv_to_stream(stream, byte_mv, cursor):
    """
    Write numpy byte array view to stream
    - use select to handle forward pressure
    - use a while loop in case the stream is non-blocking (meaning the ammount of byte written is not guarantied to be cursor len)
    Args:
        stream: stream to write to
        byte_mv: numpy byte view of the buffer to write
        cursor: ammount of byte to write

    """
    written = 0
    while written < cursor:
        _, writable, exceptional = select([], [stream], [stream])
        if exceptional:
            raise IOError(f'error with input stream, {exceptional}')
        written += stream.write(byte_mv[written:cursor].tobytes())



"""
Entry point to the run method of summarypy
read an event loss stream and aggregate it into the relevant summary loss stream

using numba latest version 0.59, we are limited is our usage of classes and have to pass each data structure into the different function
The attributes are name the same throughout the code, below is their definition

Static intermediary data structure:
    summary_pipes : dict summary_set_id to path then updated to stream when reading starts
    summary_sets_id : array of all summary_set_id that have a stream
    summary_set_id_to_summary_set_index : array mapping summary_set_id => summary_set_index
    summary_set_index_to_loss_ptr : mapping array summary_set_index => starting index of summary_id loss (first dimension) in the loss_summary table
    item_id_to_summary_id : mapping array item_id => [summary_id for each summary set]
    item_id_to_risks_i : mapping array item_id => risks_i (index of th corresponding risk)

event based intermediary data structure:
    summary_set_index_to_present_loss_ptr_end : copy of summary_set_index_to_loss_ptr that store the end of list of summary_id for each summary set
    present_summary_id : array that store all the summary_id found in the event
                         each summary set start at summary_set_index_to_loss_ptr and ends at summary_set_index_to_present_loss_ptr_end
    loss_index : store the index in loss_summary of each summary corresponding to an item for quick lookup on read
    loss_summary : 2D array of losses for each summary
                   loss_summary[loss_index[summary_set_index], sidx] = loss
    is_risk_affected : array to store in if a risk has already been affected or not in this event

"""
import numpy as np
import numba as nb
from numba.typed import Dict as nb_dict
import pandas as pd

from contextlib import ExitStack
import logging
import os
from itertools import zip_longest

from oasislmf.execution.bash import RUNTYPE_GROUNDUP_LOSS, RUNTYPE_INSURED_LOSS, RUNTYPE_REINSURANCE_LOSS
from oasislmf.pytools.common.data import (load_as_ndarray, oasis_int, nb_oasis_int, oasis_int_size, oasis_float, oasis_float_size,
                                          null_index, summary_xref_dtype)
from oasislmf.pytools.common.event_stream import (EventReader, init_streams_in, stream_info_to_bytes, write_mv_to_stream,
                                                  mv_read, mv_write_summary_header, mv_write_sidx_loss, mv_write_delimiter,
                                                  GUL_STREAM_ID, FM_STREAM_ID, LOSS_STREAM_ID, SUMMARY_STREAM_ID, ITEM_STREAM, PIPE_CAPACITY,
                                                  MEAN_IDX, TIV_IDX, NUMBER_OF_AFFECTED_RISK_IDX, MAX_LOSS_IDX)
from oasislmf.pytools.utils import redirect_logging

logger = logging.getLogger(__name__)


SPECIAL_SIDX_COUNT = 6  # 0 is included as a special sidx
SUMMARY_HEADER_SIZE = 2 * oasis_int_size + oasis_float_size + SPECIAL_SIDX_COUNT * (oasis_int_size + oasis_float_size)
SIDX_LOSS_WRITE_SIZE = 2 * (oasis_int_size + oasis_float_size)


SUPPORTED_SUMMARY_SET_ID = list(range(1, 10))

risk_key_type = nb.types.UniTuple(nb_oasis_int, 2)

summary_info_dtype = np.dtype([('nb_risk', oasis_int), ])

SUPPORTED_RUN_TYPE = [RUNTYPE_GROUNDUP_LOSS, RUNTYPE_INSURED_LOSS, RUNTYPE_REINSURANCE_LOSS]


def create_summary_object_file(static_path, run_type):
    """create and write summary object into static path"""
    summary_objects = get_summary_object(static_path, run_type)
    write_summary_objects(static_path, run_type, *summary_objects)


def load_summary_object(static_path, run_type):
    """load already prepare summary data structure if present otherwise create them"""
    if os.path.isfile(os.path.join(static_path, run_type, 'summary_info.npy')):
        return read_summary_objects(static_path, run_type)
    else:
        return get_summary_object(static_path, run_type)


def get_summary_object(static_path, run_type):
    """read static files to get summary static data structure"""

    # extract item_id to index in the loss summary
    if run_type == RUNTYPE_GROUNDUP_LOSS:
        summary_xref_csv_col_map = {'summary_set_id': 'summaryset_id'}
        summary_xref = load_as_ndarray(static_path, 'gulsummaryxref', summary_xref_dtype, col_map=summary_xref_csv_col_map)
        summary_map = pd.read_csv(os.path.join(static_path, 'gul_summary_map.csv'),
                                  usecols=['loc_id', 'item_id', 'building_id', 'coverage_id'],
                                  dtype=oasis_int)
    elif run_type == RUNTYPE_INSURED_LOSS:
        summary_xref_csv_col_map = {'summary_set_id': 'summaryset_id',
                                    'item_id': 'output'}
        summary_xref = load_as_ndarray(static_path, 'fmsummaryxref', summary_xref_dtype, col_map=summary_xref_csv_col_map)
        summary_map = pd.read_csv(os.path.join(static_path, 'fm_summary_map.csv'),
                                  usecols=['loc_id', 'output_id', 'building_id', 'coverage_id'],
                                  dtype=oasis_int,
                                  ).rename(columns={'output_id': 'item_id'})
    elif run_type == RUNTYPE_REINSURANCE_LOSS:
        summary_xref_csv_col_map = {'summary_set_id': 'summaryset_id',
                                    'item_id': 'output'}
        summary_xref = load_as_ndarray(static_path, 'fmsummaryxref', summary_xref_dtype, col_map=summary_xref_csv_col_map)
        summary_map = None  # numba use none to optimise function when some part are not used
    else:
        raise Exception(f"run type {run_type} not in supported list {SUPPORTED_RUN_TYPE}")

    summary_sets_id = np.sort(np.unique(summary_xref['summary_set_id']))
    summary_set_id_to_summary_set_index = get_summary_set_id_to_summary_set_index(summary_sets_id)
    summary_set_index_to_loss_ptr, item_id_to_summary_id = get_summary_xref_info(summary_xref, summary_sets_id, summary_set_id_to_summary_set_index)

    if summary_map is not None:
        nb_risk, item_id_to_risks_i = extract_risk_info(item_id_to_summary_id.shape[0], summary_map)
    else:
        item_id_to_risks_i = np.zeros(0, dtype=oasis_int)
        nb_risk = 0

    summary_info = np.empty(1, dtype=summary_info_dtype)
    info = summary_info[0]
    info['nb_risk'] = nb_risk

    return summary_info, summary_set_id_to_summary_set_index, summary_set_index_to_loss_ptr, item_id_to_summary_id, item_id_to_risks_i


def write_summary_objects(static_path, run_type, summary_info, summary_set_id_to_summary_set_index,
                          summary_set_index_to_loss_ptr, item_id_to_summary_id, item_id_to_risks_i):
    os.makedirs(os.path.join(static_path, run_type), exist_ok=True)
    np.save(os.path.join(static_path, run_type, 'summary_info'), summary_info)
    np.save(os.path.join(static_path, run_type, 'summary_set_id_to_summary_set_index'), summary_set_id_to_summary_set_index)
    np.save(os.path.join(static_path, run_type, 'summary_set_index_to_loss_ptr'), summary_set_index_to_loss_ptr)
    np.save(os.path.join(static_path, run_type, 'item_id_to_summary_id'), item_id_to_summary_id)
    np.save(os.path.join(static_path, run_type, 'item_id_to_risks_i'), item_id_to_risks_i)


def read_summary_objects(static_path, run_type):
    summary_info = np.load(os.path.join(static_path, run_type, 'summary_info.npy'), mmap_mode='r')
    summary_set_id_to_summary_set_index = np.load(os.path.join(static_path, run_type, 'summary_set_id_to_summary_set_index.npy'), mmap_mode='r')
    summary_set_index_to_loss_ptr = np.load(os.path.join(static_path, run_type, 'summary_set_index_to_loss_ptr.npy'), mmap_mode='r')
    item_id_to_summary_id = np.load(os.path.join(static_path, run_type, 'item_id_to_summary_id.npy'), mmap_mode='r')
    item_id_to_risks_i = np.load(os.path.join(static_path, run_type, 'item_id_to_risks_i.npy'), mmap_mode='r')
    return summary_info, summary_set_id_to_summary_set_index, summary_set_index_to_loss_ptr, item_id_to_summary_id, item_id_to_risks_i


@nb.njit(cache=True)
def nb_extract_risk_info(item_id_to_risks_i, summary_map_item_ids, summary_map_loc_ids, summary_map_building_ids):
    loc_id_building_id_to_building_risk = nb_dict.empty(risk_key_type, nb_oasis_int)
    last_risk_i = 0
    for i in range(summary_map_item_ids.shape[0]):
        loc_id_building_id = (nb_oasis_int(summary_map_loc_ids[i]), nb_oasis_int(summary_map_building_ids[i]))
        if loc_id_building_id in loc_id_building_id_to_building_risk:
            risk_i = loc_id_building_id_to_building_risk[loc_id_building_id]
        else:
            loc_id_building_id_to_building_risk[loc_id_building_id] = risk_i = nb_oasis_int(last_risk_i)
            last_risk_i += 1
        item_id_to_risks_i[summary_map_item_ids[i]] = nb_oasis_int(risk_i)
    return last_risk_i


def extract_risk_info(len_item_id, summary_map):
    """
    extract relevant information regarding item and risk mapping from summary_map
    Args:
        len_item_id: number of items
        summary_map: numpy ndarray view of the summary_map

    Returns:
        (number of risk, mapping array item_id => risks_i)
    """

    item_id_to_risks_i = np.zeros(len_item_id, oasis_int)
    nb_risk = nb_extract_risk_info(
        item_id_to_risks_i,
        summary_map['item_id'].astype(oasis_int).to_numpy(),
        summary_map['loc_id'].astype(oasis_int).to_numpy(),
        summary_map['building_id'].astype(oasis_int).to_numpy())
    return nb_risk, item_id_to_risks_i


@nb.jit(nopython=True, cache=True)
def read_buffer(byte_mv, cursor, valid_buff, event_id, item_id,
                summary_sets_id, summary_set_index_to_loss_ptr, item_id_to_summary_id,
                loss_index, loss_summary, present_summary_id, summary_set_index_to_present_loss_ptr_end,
                item_id_to_risks_i, is_risk_affected, has_affected_risk):
    """read valid part of byte_mv and load relevant data for one event"""
    last_event_id = event_id
    while True:
        if item_id:
            if valid_buff - cursor < (oasis_int_size + oasis_float_size):
                break
            sidx, cursor = mv_read(byte_mv, cursor, oasis_int, oasis_int_size)
            if sidx:
                loss, cursor = mv_read(byte_mv, cursor, oasis_float, oasis_float_size)
                loss = 0 if np.isnan(loss) else loss

                ###### do loss read ######
                if sidx > 0 or sidx in [-1, -3, -5]:
                    for summary_set_index in range(summary_sets_id.shape[0]):
                        loss_summary[loss_index[summary_set_index], sidx] += loss
                ##########

            else:
                ##### do item exit ####
                ##########
                cursor += oasis_float_size
                item_id = 0
        else:
            if valid_buff - cursor < 2 * oasis_int_size:
                break
            event_id, cursor = mv_read(byte_mv, cursor, oasis_int, oasis_int_size)
            if event_id != last_event_id:
                if last_event_id:  # we have a new event we return the one we just finished
                    for summary_set_index in range(summary_sets_id.shape[0]):  # reorder summary_id for each summary set
                        summary_set_start = summary_set_index_to_loss_ptr[summary_set_index]
                        summary_set_end = summary_set_index_to_present_loss_ptr_end[summary_set_index]
                        present_summary_id[summary_set_start: summary_set_end] = np.sort(present_summary_id[summary_set_start: summary_set_end])
                    return cursor - oasis_int_size, last_event_id, 0, 1
                else:  # first pass we store the event we are reading
                    last_event_id = event_id
            item_id, cursor = mv_read(byte_mv, cursor, oasis_int, oasis_int_size)

            ##### do new item setup #####
            if has_affected_risk is not None:
                risk_i = item_id_to_risks_i[item_id]
                if is_risk_affected[risk_i]:
                    new_risk = 0
                else:
                    new_risk = is_risk_affected[risk_i] = 1

            for summary_set_index in range(summary_sets_id.shape[0]):
                loss_index[summary_set_index] = nb_oasis_int(summary_set_index_to_loss_ptr[summary_set_index]
                                                             + item_id_to_summary_id[item_id, summary_set_index] - 1)
                # print(summary_set_index_to_loss_ptr[summary_set_index], item_id_to_summary_id[item_id, summary_set_index])
                # print('new_risk', event_id, item_id, new_risk)
                # print('loss_index[summary_set_index]', loss_index[summary_set_index])
                # print('loss_summary[loss_index[summary_set_index], NUMBER_OF_AFFECTED_RISK_IDX]', loss_summary[loss_index[summary_set_index], -4])
                if loss_summary[loss_index[summary_set_index], NUMBER_OF_AFFECTED_RISK_IDX] == 0:  # we use sidx 0 to check if this summary_id has already been seen
                    present_summary_id[summary_set_index_to_present_loss_ptr_end[summary_set_index]
                                       ] = item_id_to_summary_id[item_id, summary_set_index]
                    summary_set_index_to_present_loss_ptr_end[summary_set_index] += 1
                    loss_summary[loss_index[summary_set_index], NUMBER_OF_AFFECTED_RISK_IDX] = 1

                elif has_affected_risk is not None:
                    loss_summary[loss_index[summary_set_index], NUMBER_OF_AFFECTED_RISK_IDX] += new_risk
            ##########
    for summary_set_index in range(summary_sets_id.shape[0]):  # reorder summary_id for each summary set
        summary_set_start = summary_set_index_to_loss_ptr[summary_set_index]
        summary_set_end = summary_set_index_to_present_loss_ptr_end[summary_set_index]
        present_summary_id[summary_set_start: summary_set_end] = np.sort(present_summary_id[summary_set_start: summary_set_end])
    return cursor, event_id, item_id, 0


@nb.njit(cache=True)
def mv_write_event(byte_mv, event_id, len_sample, last_loss_summary_index, last_sidx,
                   output_zeros, has_affected_risk,
                   summary_set_index, summary_set_index_to_loss_ptr, summary_set_index_to_present_loss_ptr_end, present_summary_id, loss_summary,
                   summary_index_cursor, summary_sets_cursor, summary_stream_index):
    """
        load event summary loss into byte_mv

    Args:
        byte_mv: numpy byte view to write to the stream
        event_id: event id
        len_sample: max sample id
        last_loss_summary_index: last summary index written (used to restart from the last summary when buffer was full)
        last_sidx: last sidx written in the buffer (used to restart from the correct sidx when buffer was full

        see other args definition in run method
    """
    cursor = 0
    for loss_summary_index in range(max(summary_set_index_to_loss_ptr[summary_set_index], last_loss_summary_index),
                                    summary_set_index_to_present_loss_ptr_end[summary_set_index]):
        summary_id = present_summary_id[loss_summary_index]
        losses = loss_summary[summary_set_index_to_loss_ptr[summary_set_index] + summary_id - 1]

        if not output_zeros and losses[TIV_IDX] == 0 and losses[MEAN_IDX] == 0:
            continue

        summary_stream_index[summary_index_cursor]['summary_id'] = summary_id
        # we use offset to temporally store the cursor, we set the correct value later on
        summary_stream_index[summary_index_cursor]['offset'] = cursor

        if last_sidx == 0:
            if cursor < PIPE_CAPACITY - SUMMARY_HEADER_SIZE:
                cursor = mv_write_summary_header(byte_mv, cursor, event_id, summary_id, losses[TIV_IDX])
                cursor = mv_write_sidx_loss(byte_mv, cursor, MEAN_IDX, losses[MEAN_IDX])
                if has_affected_risk is not None:
                    cursor = mv_write_sidx_loss(byte_mv, cursor, NUMBER_OF_AFFECTED_RISK_IDX, losses[NUMBER_OF_AFFECTED_RISK_IDX])
                cursor = mv_write_sidx_loss(byte_mv, cursor, MAX_LOSS_IDX, losses[MAX_LOSS_IDX])
                last_sidx = 1
            else:
                return cursor, loss_summary_index, last_sidx, summary_index_cursor

        for sidx in range(last_sidx, len_sample + 1):
            if not output_zeros and losses[sidx] == 0:
                continue
            if cursor < PIPE_CAPACITY - SIDX_LOSS_WRITE_SIZE:  # times 2 to accommodate 0,0 if last item
                cursor = mv_write_sidx_loss(byte_mv, cursor, sidx, losses[sidx])
            else:
                return cursor, loss_summary_index, sidx, summary_index_cursor

        cursor = mv_write_delimiter(byte_mv, cursor)
        # set the correct offset for idx file and update summary_sets_cursor
        summary_byte_len = cursor - summary_stream_index[summary_index_cursor]['offset']
        summary_stream_index[summary_index_cursor]['offset'] = summary_sets_cursor[summary_set_index]
        summary_sets_cursor[summary_set_index] += summary_byte_len
        summary_index_cursor += 1

        last_sidx = 0
    return cursor, -1, 0, summary_index_cursor


class SummaryReader(EventReader):
    def __init__(self, summary_sets_id, summary_set_index_to_loss_ptr, item_id_to_summary_id,
                 loss_index, loss_summary, present_summary_id, summary_set_index_to_present_loss_ptr_end,
                 item_id_to_risks_i, is_risk_affected, has_affected_risk):
        self.summary_sets_id = summary_sets_id
        self.summary_set_index_to_loss_ptr = summary_set_index_to_loss_ptr
        self.item_id_to_summary_id = item_id_to_summary_id
        self.loss_index = loss_index
        self.loss_summary = loss_summary
        self.present_summary_id = present_summary_id
        self.summary_set_index_to_present_loss_ptr_end = summary_set_index_to_present_loss_ptr_end
        self.item_id_to_risks_i = item_id_to_risks_i
        self.is_risk_affected = is_risk_affected
        self.has_affected_risk = has_affected_risk
        self.logger = logger

    def read_buffer(self, byte_mv, cursor, valid_buff, event_id, item_id):
        return read_buffer(
            byte_mv, cursor, valid_buff, event_id, item_id,
            self.summary_sets_id, self.summary_set_index_to_loss_ptr, self.item_id_to_summary_id,
            self.loss_index, self.loss_summary, self.present_summary_id, self.summary_set_index_to_present_loss_ptr_end,
            self.item_id_to_risks_i, self.is_risk_affected, self.has_affected_risk
        )


def get_summary_set_id_to_summary_set_index(summary_sets_id):
    """create an array mapping summary_set_id => summary_set_index"""
    summary_set_id_to_summary_set_index = np.full(np.max(summary_sets_id) + 1, null_index, 'i4')
    for summary_set_index in range(summary_sets_id.shape[0]):
        summary_set_id_to_summary_set_index[summary_sets_id[summary_set_index]] = summary_set_index
    return summary_set_id_to_summary_set_index


def get_summary_xref_info(summary_xref, summary_sets_id, summary_set_id_to_summary_set_index):
    """
    extract mapping from summary_xref
    """
    summary_set_index_to_loss_ptr = np.zeros(summary_sets_id.shape[0] + 1, oasis_int)
    max_item_id = 0
    for i in range(summary_xref.shape[0]):
        xref = summary_xref[i]
        summary_set_index = summary_set_id_to_summary_set_index[xref['summary_set_id']]
        if summary_set_index == null_index:
            continue
        if xref['summary_id'] > summary_set_index_to_loss_ptr[summary_set_index + 1]:
            summary_set_index_to_loss_ptr[summary_set_index + 1] = xref['summary_id']
        if xref['item_id'] > max_item_id:
            max_item_id = xref['item_id']
    for i in range(1, summary_set_index_to_loss_ptr.shape[0]):
        summary_set_index_to_loss_ptr[i] += summary_set_index_to_loss_ptr[i - 1]

    item_id_to_summary_id = np.full((max_item_id + 1, summary_sets_id.shape[0]), null_index, oasis_int)
    for i in range(summary_xref.shape[0]):
        xref = summary_xref[i]
        summary_set_index = summary_set_id_to_summary_set_index[xref['summary_set_id']]
        if summary_set_index == null_index:
            continue
        item_id_to_summary_id[xref['item_id'], summary_set_index] = xref['summary_id']

    return summary_set_index_to_loss_ptr, item_id_to_summary_id


def run(files_in, static_path, run_type, low_memory, output_zeros, **kwargs):
    """
    Args:
        files_in: list of file path to read event from
        run_type: type of the source that is sending the stream
        static_path: path to the static files
        low_memory: if true output summary index file
        output_zeros: if true output 0 loss
        **kwargs:

    """
    summary_sets_path = {}
    error_msg = (f"summary_sets_output expected format is a list of -summary_set_id summary_set_path (ex: -1 S1.bin -2 S2.bin')"
                 f", found '{' '.join(kwargs['summary_sets_output'])}'")
    for summary_set_id, summary_set_path in zip_longest(*[iter(kwargs['summary_sets_output'])] * 2):
        if summary_set_id[0] != '-' or summary_set_path is None:
            raise Exception(error_msg)
        try:
            summary_sets_path[int(summary_set_id[1:])] = summary_set_path
        except ValueError:
            raise Exception(error_msg)
    summary_sets_id = np.array(list(summary_sets_path.keys()))

    with ExitStack() as stack:
        summary_sets_pipe = {i: stack.enter_context(open(summary_set_path, 'wb')) for i, summary_set_path in summary_sets_path.items()}

        if low_memory:
            summary_sets_index_pipe = {summary_set_id: stack.enter_context(open(setpath.rsplit('.', 1)[0] + '.idx', 'w'))
                                       for summary_set_id, setpath in summary_sets_path.items()}

        streams_in, (stream_source_type, stream_agg_type, len_sample) = init_streams_in(files_in, stack)

        if stream_source_type not in (GUL_STREAM_ID, FM_STREAM_ID, LOSS_STREAM_ID):
            raise Exception(f"unsupported stream type {stream_source_type}, {stream_agg_type}")

        summary_object = load_summary_object(static_path, run_type)
        summary_info, summary_set_id_to_summary_set_index, summary_set_index_to_loss_ptr, item_id_to_summary_id, item_id_to_risks_i = summary_object

        # check all summary_set_id required are defined in summaryxref
        invalid_summary_sets_id = []
        for summary_set_id in summary_sets_id:
            if summary_set_id > summary_set_id_to_summary_set_index.shape[0] or summary_set_id_to_summary_set_index[summary_set_id] == null_index:
                invalid_summary_sets_id.append(summary_set_id)
        if invalid_summary_sets_id:
            raise ValueError(
                f'summary_set_ids {invalid_summary_sets_id} not found in summaryxref available summary_set_id '
                f'{[summary_set_id for summary_set_id, summary_set_index in enumerate(summary_set_id_to_summary_set_index) if summary_set_index != null_index]}')

        has_affected_risk = True if summary_info[0]['nb_risk'] > 0 else None

        # init temporary
        present_summary_id = np.zeros(summary_set_index_to_loss_ptr[-1], oasis_int)
        loss_index = np.empty(summary_sets_id.shape[0], oasis_int)
        summary_set_index_to_present_loss_ptr_end = np.array(summary_set_index_to_loss_ptr)
        loss_summary = np.zeros((summary_set_index_to_loss_ptr[-1], len_sample + SPECIAL_SIDX_COUNT), dtype=oasis_float)
        is_risk_affected = np.zeros(summary_info[0]['nb_risk'], dtype=oasis_int)

        summary_reader = SummaryReader(
            summary_sets_id, summary_set_index_to_loss_ptr, item_id_to_summary_id,
            loss_index, loss_summary, present_summary_id, summary_set_index_to_present_loss_ptr_end,
            item_id_to_risks_i, is_risk_affected, has_affected_risk)

        out_byte_mv = np.frombuffer(buffer=memoryview(bytearray(PIPE_CAPACITY)), dtype='b')

        # data for index file (low_memory==True)
        summary_sets_cursor = np.zeros(summary_sets_id.shape[0], dtype=np.int64)
        summary_stream_index = np.empty(summary_set_index_to_loss_ptr[-1], dtype=np.dtype([('summary_id', oasis_int), ('offset', np.int64)]))

        for summary_set_index, summary_set_id in enumerate(summary_sets_id):
            summary_pipe = summary_sets_pipe[summary_set_id]
            summary_sets_cursor[summary_set_index] += summary_pipe.write(stream_info_to_bytes(SUMMARY_STREAM_ID, ITEM_STREAM))
            summary_sets_cursor[summary_set_index] += summary_pipe.write(len_sample.tobytes())
            summary_sets_cursor[summary_set_index] += summary_pipe.write(nb_oasis_int(summary_set_id).tobytes())

        for event_id in summary_reader.read_streams(streams_in):
            for summary_set_index, summary_set_id in enumerate(summary_sets_id):
                summary_pipe = summary_sets_pipe[summary_set_id]
                summary_index_cursor = 0
                last_loss_summary_index = 0
                last_sidx = 0
                while True:
                    cursor, last_loss_summary_index, last_sidx, summary_index_cursor = mv_write_event(
                        out_byte_mv, event_id, len_sample, last_loss_summary_index, last_sidx,
                        output_zeros, has_affected_risk,
                        summary_set_index, summary_set_index_to_loss_ptr, summary_set_index_to_present_loss_ptr_end, present_summary_id,
                        loss_summary,
                        summary_index_cursor, summary_sets_cursor, summary_stream_index
                    )
                    write_mv_to_stream(summary_pipe, out_byte_mv, cursor)
                    if last_loss_summary_index == -1:
                        break
                if low_memory:
                    # write the summary.idx file
                    np.savetxt(summary_sets_index_pipe[summary_set_id],
                               summary_stream_index[:summary_index_cursor],
                               fmt="%i,%i")

            loss_summary.fill(0)
            present_summary_id.fill(0)
            summary_set_index_to_present_loss_ptr_end[:] = summary_set_index_to_loss_ptr
            is_risk_affected.fill(0)


@redirect_logging(exec_name='summarypy')
def main(create_summarypy_files, static_path, run_type, **kwargs):
    if create_summarypy_files:
        create_summary_object_file(static_path, run_type)
    else:
        run(static_path=static_path, run_type=run_type, **kwargs)



__all__ = [
    'GenerateLosses',
    'GenerateLossesDir',
    'GenerateLossesPartial',
    'GenerateLossesOutput',
    'GenerateLossesDeterministic',
    'GenerateLossesDummyModel'
]

import importlib
import io
import json
import multiprocessing
import os
import re
import subprocess
import sys
import warnings
from collections import OrderedDict
from json import JSONDecodeError
from pathlib import Path
from subprocess import CalledProcessError, check_call

import pandas as pd
import numpy as np

from oasis_data_manager.filestore.config import get_storage_from_config_path
from oasis_data_manager.filestore.backends.local import LocalStorage

from ods_tools.oed.setting_schema import AnalysisSettingSchema, ModelSettingSchema

from ...execution import bash, runner
from ...execution.bash import get_fmcmd, RUNTYPE_GROUNDUP_LOSS, RUNTYPE_INSURED_LOSS, RUNTYPE_REINSURANCE_LOSS
from ...execution.bin import (csv_to_bin, prepare_run_directory,
                              prepare_run_inputs, set_footprint_set, set_vulnerability_set, set_loss_factors_set)
from ...preparation.summaries import generate_summaryxref_files
from ...pytools.fm.financial_structure import create_financial_structure
from oasislmf.pytools.summary.manager import create_summary_object_file
from ...utils.data import (get_dataframe, get_exposure_data, get_json,
                           get_utctimestamp, merge_dataframes, set_dataframe_column_dtypes)
from ...utils.defaults import (EVE_DEFAULT_SHUFFLE, EVE_STD_SHUFFLE, KTOOL_N_FM_PER_LB,
                               KTOOL_N_GUL_PER_LB, KTOOLS_ALLOC_FM_MAX, KTOOLS_ALLOC_GUL_DEFAULT,
                               KTOOLS_ALLOC_GUL_MAX, KTOOLS_ALLOC_IL_DEFAULT,
                               KTOOLS_ALLOC_RI_DEFAULT, KTOOLS_DEBUG, KTOOLS_GUL_LEGACY_STREAM,
                               KTOOLS_MEAN_SAMPLE_IDX, KTOOLS_NUM_PROCESSES,
                               KTOOLS_STD_DEV_SAMPLE_IDX, KTOOLS_TIV_SAMPLE_IDX)
from ...utils.exceptions import OasisException
from ...utils.inputs import str2bool
from ...utils.path import setcwd
from ..base import ComputationStep
from .files import GenerateDummyModelFiles, GenerateDummyOasisFiles

warnings.simplefilter(action='ignore', category=FutureWarning)


class GenerateLossesBase(ComputationStep):
    """
    Base class for Loss generation functions

    Includes methods useful across all GenerateLoss functions
    intended as a common inherited class
    """

    def _get_output_dir(self):
        """
        Set the model run directory to '<cwd>/runs/losses-<timestamp>' if not set
        in arguments

        :return: (str) the model run directory, either given or generated
        """
        if not self.model_run_dir:
            utcnow = get_utctimestamp(fmt='%Y%m%d%H%M%S')
            self.model_run_dir = os.path.join(os.getcwd(), 'runs', 'losses-{}'.format(utcnow))
        if not os.path.exists(self.model_run_dir):
            Path(self.model_run_dir).mkdir(parents=True, exist_ok=True)
        return self.model_run_dir

    def _get_model_runner(self):
        """
        Returns the model runner module, by default this is imported from `oasislmf/execution/runner.py`
        but can be overridden from the conf option `model_package_dir`

        :return: (object) The model runner module, (str) Package Name
        """
        package_name = None
        if self.model_package_dir and os.path.exists(os.path.join(self.model_package_dir, 'supplier_model_runner.py')):
            path, package_name = os.path.split(self.model_package_dir)
            sys.path.append(path)
            return importlib.import_module('{}.supplier_model_runner'.format(package_name)), package_name
        else:
            return runner, package_name

    def _check_ktool_rules(self):
        """
        Check the given ktool allocation rules are within valid ranges
        Raises an `OasisException` if a rules is invalid
        """
        rule_ranges = {
            'ktools_alloc_rule_gul': KTOOLS_ALLOC_GUL_MAX,
            'ktools_alloc_rule_il': KTOOLS_ALLOC_FM_MAX,
            'ktools_alloc_rule_ri': KTOOLS_ALLOC_FM_MAX,
            'ktools_event_shuffle': EVE_STD_SHUFFLE,
            'gulpy_random_generator': 1}

        for rule in rule_ranges:
            rule_val = int(getattr(self, rule))
            if (rule_val < 0) or (rule_val > rule_ranges[rule]):
                raise OasisException(f'Error: {rule}={rule_val} - Not within valid ranges [0..{rule_ranges[rule]}]')

    def _store_run_settings(self, analysis_settings, target_dir):
        """
        Writes the analysis settings file to the `target_dir` path
        """
        with io.open(os.path.join(target_dir, 'analysis_settings.json'), 'w', encoding='utf-8') as f:
            f.write(json.dumps(analysis_settings, ensure_ascii=False, indent=4))

    def _get_num_ri_layers(self, analysis_settings, model_run_fp):
        """
        Find the number of Reinsurance layers based on `'ri_layers.json'`, returns pos int()
        """
        ri_layers = 0
        if analysis_settings.get('ri_output', False) or analysis_settings.get('rl_output', False):
            ri_layers = len(get_json(os.path.join(model_run_fp, 'input', 'ri_layers.json')))
        return ri_layers

    def _get_peril_filter(self, analysis_settings):
        """
        Check the 'analysis_settings' for user set peril filter, if empty return the MDK peril filter
        option
        """
        user_peril_filter = analysis_settings.get('peril_filter', None)
        peril_filter = list(map(str.upper, user_peril_filter if user_peril_filter else self.peril_filter))
        return peril_filter

    def _print_error_logs(self, run_log_fp, e):
        """
        Error handling Method: Call if a run error has accursed,
        * prints ktool log files to logger
        * Raises `OasisException`
        """
        bash_trace_fp = os.path.join(run_log_fp, 'bash.log')
        if os.path.isfile(bash_trace_fp):
            with io.open(bash_trace_fp, 'r', encoding='utf-8') as f:
                self.logger.info('\nBASH_TRACE:\n' + "".join(f.readlines()))

        stderror_fp = os.path.join(run_log_fp, 'stderror.err')
        if os.path.isfile(stderror_fp):
            with io.open(stderror_fp, 'r', encoding='utf-8') as f:
                self.logger.info('\nKTOOLS_STDERR:\n' + "".join(f.readlines()))

        gul_stderror_fp = os.path.join(run_log_fp, 'gul_stderror.err')
        if os.path.isfile(gul_stderror_fp):
            with io.open(gul_stderror_fp, 'r', encoding='utf-8') as f:
                self.logger.info('\nGUL_STDERR:\n' + "".join(f.readlines()))

        self.logger.info('\nSTDOUT:\n' + e.output.decode('utf-8').strip())

        raise OasisException(
            'Ktools run Error: non-zero exit code or error/warning messages detected in STDERR output.\n'
            'Killing all processes. To disable this automated check run with `--ktools-disable-guard`.\n'
            'Logs stored in: {}'.format(run_log_fp)
        )


class GenerateLossesDir(GenerateLossesBase):
    """
    Prepare the loss generation directory

    * converts input `csv` files to ktools binary types
    * links model data to the static directory to the run locations
    * Validates and updates the `analysis_settings.json`
    * Stores the analysis_settings.json in the output directory

    :return: (dict) Updated analysis_settings
    """
    step_params = [
        # Command line options
        {'name': 'oasis_files_dir', 'flag': '-o', 'is_path': True, 'pre_exist': True,
         'required': True, 'help': 'Path to the directory in which to generate the Oasis files'},
        {'name': 'check_oed', 'type': str2bool, 'const': True, 'nargs': '?', 'default': True, 'help': 'if True check input oed files'},
        {'name': 'analysis_settings_json', 'flag': '-a', 'is_path': True, 'pre_exist': True, 'required': True,
         'help': 'Analysis settings JSON file path'},
        {'name': 'model_storage_json', 'is_path': True, 'pre_exist': True, 'required': False,
         'help': 'Model data storage settings JSON file path'},
        {'name': 'model_settings_json', 'flag': '-M', 'is_path': True, 'pre_exist': False, 'required': False,
         'help': 'Model settings JSON file path'},
        {'name': 'user_data_dir', 'flag': '-D', 'is_path': True, 'pre_exist': False,
         'help': 'Directory containing additional model data files which varies between analysis runs'},
        {'name': 'model_data_dir', 'flag': '-d', 'is_path': True, 'pre_exist': True, 'help': 'Model data directory path'},
        {'name': 'copy_model_data', 'default': False, 'type': str2bool, 'help': 'Copy model data instead of creating symbolic links to it.'},
        {'name': 'model_run_dir', 'flag': '-r', 'is_path': True, 'pre_exist': False, 'help': 'Model run directory path'},
        {'name': 'model_package_dir', 'flag': '-p', 'is_path': True, 'pre_exist': False, 'help': 'Path containing model specific package'},
        {'name': 'ktools_legacy_stream', 'type': str2bool, 'const': True, 'nargs': '?', 'default': KTOOLS_GUL_LEGACY_STREAM,
         'help': 'Run Ground up losses using the older stream type (Compatibility option)'},
        {'name': 'fmpy', 'default': True, 'type': str2bool, 'const': True, 'nargs': '?', 'help': 'use fmcalc python version instead of c++ version'},
        {'name': 'ktools_alloc_rule_il', 'default': KTOOLS_ALLOC_IL_DEFAULT, 'type': int,
         'help': 'Set the fmcalc allocation rule used in direct insured loss'},
        {'name': 'ktools_alloc_rule_ri', 'default': KTOOLS_ALLOC_RI_DEFAULT, 'type': int,
         'help': 'Set the fmcalc allocation rule used in reinsurance'},
        {'name': 'summarypy', 'default': False, 'type': str2bool, 'const': True,
            'nargs': '?', 'help': 'use summarycalc python version instead of c++ version'},
        {'name': 'check_missing_inputs', 'default': False, 'type': str2bool, 'const': True, 'nargs': '?',
         'help': 'Fail an analysis run if IL/RI is requested without the required generated files.'},

        # Manager only options (pass data directy instead of filepaths)
        {'name': 'verbose', 'default': KTOOLS_DEBUG},

    ]

    def _get_storage_manager(self):
        model_storage = get_storage_from_config_path(
            self.model_storage_json,
            os.path.join(self.model_run_dir, "static"),
        )

        # if not local test the connection to remote storage FS
        if not isinstance(model_storage, LocalStorage):
            try:
                model_storage.listdir()
            except Exception as e:
                raise OasisException('Error: Storage Manager connection issue', e)

        return model_storage

    def __check_for_parquet_output(self, analysis_settings, runtypes):
        """
        Private method to check whether ktools components were linked to third
        party parquet libraries during compilation if user requests parquet
        output.
        """
        for runtype in runtypes:
            for summary in analysis_settings.get(f'{runtype}_summaries', {}):
                if summary.get('ord_output', {}).get('parquet_format'):
                    katparquet_output = subprocess.run(
                        ['katparquet', '-v'],
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE
                    )
                    if not 'Parquet output enabled' in katparquet_output.stderr.decode():
                        raise OasisException(
                            'Parquet output format requested but not supported by ktools components. '
                            'Please set "parquet_format" to false in analysis settings file.'
                        )
                    return  # Only need to find a single request

    def __check_summary_group_support(self, analysis_settings, runtypes):
        """
        Private method to check the max number of summary groups selected.
        If that value is greater than 9 then ktools 'summarycalc' will crash.

        Stop execution if the summarypy flag is not set as True.
        """
        for runtype in runtypes:
            summary_num = len(analysis_settings.get(f'{runtype}_summaries', []))
            if summary_num > 9 and not self.summarypy:
                raise OasisException(
                    'More than 9 summaries groups are not supported in summarycalc.'
                    f'\nEither enable summarypy or reduce the number of groups set in "{runtype}_summaries".'
                    '\nThis can be set using the flag "--summarypy True", or by setting `"summarypy": True` in the oasislmf.json file.'
                )

    def run(self):
        # need to load from exposure data info or recreate it
        model_run_fp = self._get_output_dir()
        analysis_settings = AnalysisSettingSchema().get(self.analysis_settings_json)
        model_storage = self._get_storage_manager()

        il = all(p in os.listdir(self.oasis_files_dir) for p in [
            'fm_policytc.csv',
            'fm_profile.csv',
            'fm_programme.csv',
            'fm_xref.csv'])

        ri_dirs = [fn
                   for fn in os.listdir(self.oasis_files_dir) + os.listdir(self.model_run_dir)
                   if re.match(r"RI_\d+$", fn)
                   ]
        ril = any(ri_dirs)

        # Check for missing input files and either warn user or raise exception
        il_missing = analysis_settings.get('il_output', False) and not il
        ri_missing = analysis_settings.get('ri_output', False) and not ril
        rl_missing = analysis_settings.get('rl_output', False) and not ril
        if il_missing or ri_missing or rl_missing:
            missing_input_files = "{} are enabled in the analysis_settings without the generated input files. The 'generate-oasis-files' step should be rerun with account/reinsurance files.".format(
                [
                    "IL" * il_missing, "RI" * ri_missing, "RL" * rl_missing
                ]
            )
            if self.check_missing_inputs:
                raise OasisException(missing_input_files)
            else:
                self.logger.warn(missing_input_files)

        gul_item_stream = (not self.ktools_legacy_stream)
        ri = analysis_settings.get('ri_output', False) and ril
        rl = analysis_settings.get('rl_output', False) and ril
        self.logger.info('\nPreparing loss Generation (GUL=True, IL={}, RI={}, RL={})'.format(il, ri, rl))

        runtypes = ['gul'] + ['il'] * il + ['ri'] * ri + ['rl'] * rl
        self.__check_for_parquet_output(analysis_settings, runtypes)
        self.__check_summary_group_support(analysis_settings, runtypes)

        prepare_run_directory(
            model_run_fp,
            self.oasis_files_dir,
            self.model_data_dir,
            self.analysis_settings_json,
            user_data_dir=self.user_data_dir,
            ri=ri or rl,
            copy_model_data=self.copy_model_data,
            model_storage_config_fp=self.model_storage_json,
        )

        exposure_data = get_exposure_data(self, add_internal_col=True)
        location_df = exposure_data.location.dataframe
        account_df = exposure_data.account.dataframe if exposure_data.account else None
        generate_summaryxref_files(
            location_df,
            account_df,
            model_run_fp,
            analysis_settings,
            gul_item_stream=gul_item_stream,
            il=il,
            ri=ri,
            rl=rl,
            fmpy=self.fmpy
        )

        if not ri and not rl:
            fp = os.path.join(model_run_fp, 'input')
            csv_to_bin(fp, fp, il=il)
        else:
            contents = os.listdir(model_run_fp)
            for fp in [os.path.join(model_run_fp, fn) for fn in contents if re.match(r'RI_\d+$', fn) or re.match(r'input$', fn)]:
                csv_to_bin(fp, fp, il=True, ri=True)

        if not il:
            analysis_settings['il_output'] = False
            analysis_settings['il_summaries'] = []

        if not ri:
            analysis_settings['ri_output'] = False
            analysis_settings['ri_summaries'] = []

        if not rl:
            analysis_settings['rl_output'] = False
            analysis_settings['rl_summaries'] = []

        if not any(analysis_settings.get(output) for output in ['gul_output', 'il_output', 'ri_output']):
            raise OasisException(
                'No valid output settings in: {}'.format(self.analysis_settings_json))

        # Load default samples if not set in analysis settings
        # Additional check to avoid 0 samples being interpreted as false
        if not analysis_settings.get('number_of_samples') and analysis_settings.get('number_of_samples') != 0:
            if not self.model_settings_json:
                raise OasisException("'number_of_samples' not set in analysis_settings and no model_settings.json file provided for a default value.")

            default_model_samples = ModelSettingSchema().get(self.model_settings_json, key='model_default_samples')
            if default_model_samples == None:
                raise OasisException(
                    "'number_of_samples' not set in analysis_settings and no default value 'model_default_samples' found in model_settings file.")

            self.logger.info(f"Loaded samples from model_settings file: 'model_default_samples = {default_model_samples}'")
            analysis_settings['number_of_samples'] = default_model_samples

        prepare_run_inputs(analysis_settings, model_run_fp, model_storage, ri=ri or rl)

        optional_model_sets = {'footprint_set': set_footprint_set,
                               'vulnerability_set': set_vulnerability_set,
                               'pla_loss_factors_set': set_loss_factors_set}

        for model_set, model_setter in optional_model_sets.items():
            model_set_val = analysis_settings.get('model_settings', {}).get(model_set)
            if model_set_val:
                model_setter(model_set_val, model_run_fp)

        # Test call to create fmpy files in GenerateLossesDir
        if il and self.fmpy:
            il_target_dir = os.path.join(self.model_run_dir, 'input')
            self.logger.info(f'Creating FMPY structures (IL): {il_target_dir}')
            create_financial_structure(self.ktools_alloc_rule_il, il_target_dir)

        if (ri or rl) and self.fmpy:
            for ri_sub_dir in ri_dirs:
                ri_target_dir = os.path.join(self.model_run_dir, 'input', ri_sub_dir)
                self.logger.info(f'Creating FMPY structures (RI): {ri_target_dir}')
                create_financial_structure(self.ktools_alloc_rule_ri, ri_target_dir)

        if self.summarypy:
            for runtype in [RUNTYPE_GROUNDUP_LOSS, RUNTYPE_INSURED_LOSS, RUNTYPE_REINSURANCE_LOSS]:
                if analysis_settings.get(f'{runtype}_output'):
                    summaries = analysis_settings.get('{}_summaries'.format(runtype), [])
                    summary_sets_id = np.sort([summary['id'] for summary in summaries if 'id' in summary])
                    if summary_sets_id.shape[0]:
                        if runtype == RUNTYPE_REINSURANCE_LOSS:
                            summary_dirs = [os.path.join(self.model_run_dir, 'input', ri_sub_dir) for ri_sub_dir in ri_dirs]
                        else:
                            summary_dirs = [os.path.join(self.model_run_dir, 'input')]
                        for summary_dir in summary_dirs:
                            self.logger.info(f'Creating summarypy structures {runtype}: {summary_dir}')
                            create_summary_object_file(summary_dir, runtype)

        self._store_run_settings(analysis_settings, os.path.join(model_run_fp, 'output'))
        return analysis_settings


class GenerateLossesPartial(GenerateLossesDir):
    """
    Runs a single analysis event chunk
    """
    step_params = GenerateLossesDir.step_params + [
        {'name': 'ktools_num_processes', 'flag': '-n', 'type': int, 'default': KTOOLS_NUM_PROCESSES,
         'help': 'Number of ktools calculation processes to use'},
        {'name': 'ktools_event_shuffle', 'default': EVE_DEFAULT_SHUFFLE, 'type': int,
         'help': 'Set rule for event shuffling between eve partions, 0 - No shuffle, 1 - round robin (output elts sorted), 2 - Fisher-Yates shuffle, 3 - std::shuffle (previous default in oasislmf<1.14.0) '},
        {'name': 'ktools_alloc_rule_gul', 'default': KTOOLS_ALLOC_GUL_DEFAULT, 'type': int, 'help': 'Set the allocation used in gulcalc'},
        {'name': 'ktools_alloc_rule_il', 'default': KTOOLS_ALLOC_IL_DEFAULT, 'type': int,
         'help': 'Set the fmcalc allocation rule used in direct insured loss'},
        {'name': 'ktools_alloc_rule_ri', 'default': KTOOLS_ALLOC_RI_DEFAULT, 'type': int,
         'help': 'Set the fmcalc allocation rule used in reinsurance'},
        {'name': 'ktools_num_gul_per_lb', 'default': KTOOL_N_GUL_PER_LB, 'type': int,
         'help': 'Number of gul per load balancer (0 means no load balancer)'},
        {'name': 'ktools_num_fm_per_lb', 'default': KTOOL_N_FM_PER_LB, 'type': int,
         'help': 'Number of fm per load balancer (0 means no load balancer)'},
        {'name': 'ktools_disable_guard', 'default': False, 'type': str2bool, 'const': True, 'nargs': '?',
         'help': 'Disables error handling in the ktools run script (abort on non-zero exitcode or output on stderr)'},
        {'name': 'ktools_fifo_relative', 'default': False, 'type': str2bool, 'const': True,
         'nargs': '?', 'help': 'Create ktools fifo queues under the ./fifo dir'},
        {'name': 'modelpy', 'default': False, 'type': str2bool, 'const': True, 'nargs': '?',
         'help': 'use getmodel python version instead of c++ version'},
        {'name': 'gulpy', 'default': False, 'type': str2bool, 'const': True, 'nargs': '?',
         'help': 'use gulcalc python version instead of c++ version'},
        {'name': 'gulpy_random_generator', 'default': 1, 'type': int,
         'help': 'set the random number generator in gulpy (0: Mersenne-Twister, 1: Latin Hypercube. Default: 1).'},
        {'name': 'gulmc', 'default': True, 'type': str2bool, 'const': True, 'nargs': '?', 'help': 'use full Monte Carlo gulcalc python version'},
        {'name': 'gulmc_random_generator', 'default': 1, 'type': int,
         'help': 'set the random number generator in gulmc (0: Mersenne-Twister, 1: Latin Hypercube. Default: 1).'},
        {'name': 'gulmc_effective_damageability', 'default': False, 'type': str2bool, 'const': True, 'nargs': '?',
         'help': 'use the effective damageability to draw loss samples instead of the full Monte Carlo method. Default: False'},
        {'name': 'gulmc_vuln_cache_size', 'default': 200, 'type': int,
         'help': 'Size in MB of the cache for the vulnerability calculations. Default: 200'},
        {'name': 'fmpy', 'default': True, 'type': str2bool, 'const': True, 'nargs': '?', 'help': 'use fmcalc python version instead of c++ version'},
        {'name': 'fmpy_low_memory', 'default': False, 'type': str2bool, 'const': True, 'nargs': '?',
         'help': 'use memory map instead of RAM to store loss array (may decrease performance but reduce RAM usage drastically)'},
        {'name': 'fmpy_sort_output', 'default': False, 'type': str2bool, 'const': True, 'nargs': '?', 'help': 'order fmpy output by item_id'},
        {'name': 'model_custom_gulcalc', 'default': None, 'help': 'Custom gulcalc binary name to call in the model losses step'},
        {'name': 'peril_filter', 'default': [], 'nargs': '+', 'help': 'Peril specific run'},
        {'name': 'summarypy', 'default': False, 'type': str2bool, 'const': True,
            'nargs': '?', 'help': 'use summarycalc python version instead of c++ version'},
        {'name': 'base_df_engine', 'default': "oasis_data_manager.df_reader.reader.OasisPandasReader", 'help': 'The engine to use when loading dataframes'},
        {'name': 'exposure_df_engine', 'default': None,
            'help': 'The engine to use when loading dataframes exposure data (default: same as --base-df-engine)'},
        {'name': 'model_df_engine', 'default': None,
            'help': 'The engine to use when loading dataframes model data (default: same as --base-df-engine)'},
        {'name': 'dynamic_footprint', 'default': False,
            'help': 'Dynamic Footprint'},

        # New vars for chunked loss generation
        {'name': 'analysis_settings', 'default': None},
        {'name': 'script_fp', 'default': None},
        {'name': 'process_number', 'default': None, 'type': int, 'help': 'Partition number to run, if not set then run all in a single script'},
        {'name': 'max_process_id', 'default': -1, 'type': int, 'help': 'Max number of loss chunks, defaults to `ktools_num_processes` if not set'},
        {'name': 'ktools_fifo_queue_dir', 'default': None, 'is_path': True, 'help': 'Override the path used for fifo processing'},
    ]

    def run(self):
        GenerateLossesDir._check_ktool_rules(self)
        model_run_fp = GenerateLossesDir._get_output_dir(self)

        # distributed worker will pass in run settings as JSON, if not given load settings
        # and re-load input dir.
        if not self.analysis_settings:
            self.analysis_settings = GenerateLossesDir.run(self)

        ri_layers = self._get_num_ri_layers(self.analysis_settings, model_run_fp)
        model_runner_module, _ = self._get_model_runner()

        if not self.script_fp:
            script_name = 'run_analysis.sh' if not self.process_number else f'{self.process_number}.run_analysis.sh'
            self.script_fp = os.path.join(os.path.abspath(model_run_fp), script_name)

        if os.path.isfile(self.script_fp):
            os.remove(self.script_fp)

        bash_params = bash.bash_params(
            self.analysis_settings,
            number_of_processes=self.ktools_num_processes,
            filename=self.script_fp,
            num_reinsurance_iterations=ri_layers,
            gul_alloc_rule=self.ktools_alloc_rule_gul,
            il_alloc_rule=self.ktools_alloc_rule_il,
            ri_alloc_rule=self.ktools_alloc_rule_ri,
            num_gul_per_lb=self.ktools_num_gul_per_lb,
            num_fm_per_lb=self.ktools_num_fm_per_lb,
            bash_trace=self.verbose,
            stderr_guard=not self.ktools_disable_guard,
            gul_legacy_stream=self.ktools_legacy_stream,
            fifo_tmp_dir=not self.ktools_fifo_relative,
            custom_gulcalc_cmd=self.model_custom_gulcalc,
            gulpy=(self.gulpy and not self.model_custom_gulcalc),
            gulpy_random_generator=self.gulpy_random_generator,
            gulmc=(self.gulmc and not self.model_custom_gulcalc and not self.gulpy),
            gulmc_random_generator=self.gulmc_random_generator,
            gulmc_effective_damageability=self.gulmc_effective_damageability,
            gulmc_vuln_cache_size=self.gulmc_vuln_cache_size,
            fmpy=self.fmpy,
            fmpy_low_memory=self.fmpy_low_memory,
            fmpy_sort_output=self.fmpy_sort_output,
            event_shuffle=self.ktools_event_shuffle,
            process_number=self.process_number,
            max_process_id=self.max_process_id,
            modelpy=self.modelpy,
            peril_filter=self._get_peril_filter(self.analysis_settings),
            summarypy=self.summarypy,
            exposure_df_engine=self.exposure_df_engine or self.base_df_engine,
            model_df_engine=self.model_df_engine or self.base_df_engine,
            dynamic_footprint=self.dynamic_footprint
        )
        # Workaround test -- needs adding into bash_params
        if self.ktools_fifo_queue_dir:
            bash_params['fifo_queue_dir'] = self.ktools_fifo_queue_dir

        with setcwd(model_run_fp):
            try:
                if self.process_number:
                    self.logger.info('Generated loss Chunk {} of {} in, {}'.format(
                        bash_params['process_number'],
                        bash_params['max_process_id'],
                        model_run_fp,
                    ))
                else:
                    self.logger.info('All {} Loss chunks generated in {}'.format(bash_params['max_process_id'], model_run_fp))

                return model_runner_module.run_analysis(**bash_params)
            except CalledProcessError as e:
                log_fp = os.path.join(model_run_fp, 'log', str(bash_params.get('process_number', '')))
                self._print_error_logs(log_fp, e)
        return model_run_fp


class GenerateLossesOutput(GenerateLossesDir):
    """
    Runs the output reports generation on a set of event chunks
    """
    step_params = GenerateLossesDir.step_params + [
        {'name': 'analysis_settings_json', 'flag': '-a', 'is_path': True, 'pre_exist': True, 'required': True,
         'help': 'Analysis settings JSON file path'},
        {'name': 'ktools_num_processes', 'flag': '-n', 'type': int, 'default': KTOOLS_NUM_PROCESSES,
         'help': 'Number of ktools calculation processes to use'},
        {'name': 'ktools_disable_guard', 'default': False, 'type': str2bool, 'const': True, 'nargs': '?',
         'help': 'Disables error handling in the ktools run script (abort on non-zero exitcode or output on stderr)'},
        {'name': 'ktools_fifo_relative', 'default': False, 'type': str2bool, 'const': True,
         'nargs': '?', 'help': 'Create ktools fifo queues under the ./fifo dir'},

        # New vars for chunked loss generation
        {'name': 'analysis_settings', 'default': None},
        {'name': 'script_fp', 'default': None},
        {'name': 'remove_working_file', 'default': False, 'help': 'Delete files in the "work/" dir onces outputs have completed'},
        {'name': 'max_process_id', 'default': -1, 'type': int, 'help': 'Max number of loss chunks, defaults to `ktools_num_processes` if not set'},
    ]

    def run(self):
        model_run_fp = GenerateLossesDir._get_output_dir(self)
        if not self.analysis_settings:
            self.analysis_settings = GenerateLossesDir.run(self)

        model_runner_module, _ = self._get_model_runner()
        ri_layers = self._get_num_ri_layers(self.analysis_settings, model_run_fp)

        if not self.script_fp:
            self.script_fp = os.path.join(os.path.abspath(model_run_fp), 'run_outputs.sh')

        if os.path.isfile(self.script_fp):
            os.remove(self.script_fp)

        bash_params = bash.bash_params(
            self.analysis_settings,
            number_of_processes=self.ktools_num_processes,
            num_reinsurance_iterations=ri_layers,
            filename=self.script_fp,
            bash_trace=self.verbose,
            stderr_guard=not self.ktools_disable_guard,
            fifo_tmp_dir=not self.ktools_fifo_relative,
            remove_working_file=self.remove_working_file,
            max_process_id=self.max_process_id,
        )
        with setcwd(model_run_fp):
            try:
                self.logger.info('Generating Loss outputs in {}'.format(model_run_fp))
                return model_runner_module.run_outputs(**bash_params)
            except CalledProcessError as e:
                log_fp = os.path.join(model_run_fp, 'log', 'out')
                self._print_error_logs(log_fp, e)
        return model_run_fp


class GenerateLosses(GenerateLossesDir):
    """
    Runs the GenerateLosses workflow as a single bash script (Default for the MDK)

    Generates losses using the installed ktools framework given Oasis files,
    model analysis settings JSON file, model data and model package data.

    The command line arguments can be supplied in the configuration file
    (``oasislmf.json`` by default or specified with the ``--config`` flag).
    Run ``oasislmf config --help`` for more information.

    The script creates a time-stamped folder in the model run directory and
    sets that as the new model run directory, copies the analysis settings
    JSON file into the run directory and creates the following folder
    structure
    ::

        |-- analysis_settings.json
        |-- fifo
        |-- input
            |-- RI_1
        |-- output
        |-- ri_layers.json
        |-- run_ktools.sh
        |-- static
        `-- work

    Depending on the OS type the model data is symlinked (Linux, Darwin) or
    copied (Cygwin, Windows) into the ``static`` subfolder. The input files
    are kept in the ``input`` subfolder and the losses are generated as CSV
    files in the ``output`` subfolder.
    """
    step_params = GenerateLossesDir.step_params + [
        {'name': 'ktools_num_processes', 'flag': '-n', 'type': int, 'default': KTOOLS_NUM_PROCESSES,
         'help': 'Number of ktools calculation processes to use'},
        {'name': 'ktools_event_shuffle', 'default': EVE_DEFAULT_SHUFFLE, 'type': int,
         'help': 'Set rule for event shuffling between eve partions, 0 - No shuffle, 1 - round robin (output elts sorted), 2 - Fisher-Yates shuffle, 3 - std::shuffle (previous default in oasislmf<1.14.0) '},
        {'name': 'ktools_alloc_rule_gul', 'default': KTOOLS_ALLOC_GUL_DEFAULT, 'type': int, 'help': 'Set the allocation used in gulcalc'},
        {'name': 'ktools_alloc_rule_il', 'default': KTOOLS_ALLOC_IL_DEFAULT, 'type': int,
         'help': 'Set the fmcalc allocation rule used in direct insured loss'},
        {'name': 'ktools_alloc_rule_ri', 'default': KTOOLS_ALLOC_RI_DEFAULT, 'type': int,
         'help': 'Set the fmcalc allocation rule used in reinsurance'},
        {'name': 'ktools_num_gul_per_lb', 'default': KTOOL_N_GUL_PER_LB, 'type': int,
         'help': 'Number of gul per load balancer (0 means no load balancer)'},
        {'name': 'ktools_num_fm_per_lb', 'default': KTOOL_N_FM_PER_LB, 'type': int,
         'help': 'Number of fm per load balancer (0 means no load balancer)'},
        {'name': 'ktools_disable_guard', 'default': False, 'type': str2bool, 'const': True, 'nargs': '?',
         'help': 'Disables error handling in the ktools run script (abort on non-zero exitcode or output on stderr)'},
        {'name': 'ktools_fifo_relative', 'default': False, 'type': str2bool, 'const': True,
         'nargs': '?', 'help': 'Create ktools fifo queues under the ./fifo dir'},
        {'name': 'modelpy', 'default': False, 'type': str2bool, 'const': True, 'nargs': '?',
         'help': 'use getmodel python version instead of c++ version'},
        {'name': 'gulpy', 'default': False, 'type': str2bool, 'const': True, 'nargs': '?',
         'help': 'use gulcalc python version instead of c++ version'},
        {'name': 'gulpy_random_generator', 'default': 1, 'type': int,
         'help': 'set the random number generator in gulpy (0: Mersenne-Twister, 1: Latin Hypercube. Default: 1).'},
        {'name': 'gulmc', 'default': True, 'type': str2bool, 'const': True, 'nargs': '?', 'help': 'use full Monte Carlo gulcalc python version'},
        {'name': 'gulmc_random_generator', 'default': 1, 'type': int,
         'help': 'set the random number generator in gulmc (0: Mersenne-Twister, 1: Latin Hypercube. Default: 1).'},
        {'name': 'gulmc_effective_damageability', 'default': False, 'type': str2bool, 'const': True, 'nargs': '?',
         'help': 'use the effective damageability to draw loss samples instead of the full Monte Carlo method. Default: False'},
        {'name': 'gulmc_vuln_cache_size', 'default': 200, 'type': int,
         'help': 'Size in MB of the cache for the vulnerability calculations. Default: 200'},
        {'name': 'fmpy', 'default': True, 'type': str2bool, 'const': True, 'nargs': '?', 'help': 'use fmcalc python version instead of c++ version'},
        {'name': 'fmpy_low_memory', 'default': False, 'type': str2bool, 'const': True, 'nargs': '?',
         'help': 'use memory map instead of RAM to store loss array (may decrease performance but reduce RAM usage drastically)'},
        {'name': 'fmpy_sort_output', 'default': False, 'type': str2bool, 'const': True, 'nargs': '?', 'help': 'order fmpy output by item_id'},
        {'name': 'model_custom_gulcalc', 'default': None, 'help': 'Custom gulcalc binary name to call in the model losses step'},
        {'name': 'model_py_server', 'default': False, 'type': str2bool, 'help': 'running the data server for modelpy'},
        {'name': 'peril_filter', 'default': [], 'nargs': '+', 'help': 'Peril specific run'},
        {'name': 'summarypy', 'default': False, 'type': str2bool, 'const': True,
            'nargs': '?', 'help': 'use summarycalc python version instead of c++ version'},
        {'name': 'model_custom_gulcalc_log_start', 'default': None, 'help': 'Log message produced when custom gulcalc binary process starts'},
        {'name': 'model_custom_gulcalc_log_finish', 'default': None, 'help': 'Log message produced when custom gulcalc binary process ends'},
        {'name': 'base_df_engine', 'default': "oasis_data_manager.df_reader.reader.OasisPandasReader", 'help': 'The engine to use when loading dataframes'},
        {'name': 'model_df_engine', 'default': None,
            'help': 'The engine to use when loading model data dataframes (default: --base-df-engine if not set)'},
        {'name': 'exposure_df_engine', 'default': None,
            'help': 'The engine to use when loading exposure data dataframes (default: --base-df-engine if not set)'},
        {'name': 'dynamic_footprint', 'default': False,
            'help': 'Dynamic Footprint'},
    ]

    def run(self):
        # prep losses run dir / Setup
        GenerateLossesDir._check_ktool_rules(self)
        model_run_fp = GenerateLossesDir._get_output_dir(self)
        analysis_settings = GenerateLossesDir.run(self)
        script_fp = os.path.join(os.path.abspath(model_run_fp), 'run_ktools.sh')
        ri_layers = self._get_num_ri_layers(analysis_settings, model_run_fp)
        model_runner_module, package_name = self._get_model_runner()

        with setcwd(model_run_fp):
            try:
                try:
                    model_runner_module.run(
                        analysis_settings,
                        number_of_processes=self.ktools_num_processes,
                        filename=script_fp,
                        num_reinsurance_iterations=ri_layers,
                        set_alloc_rule_gul=self.ktools_alloc_rule_gul,
                        set_alloc_rule_il=self.ktools_alloc_rule_il,
                        set_alloc_rule_ri=self.ktools_alloc_rule_ri,
                        num_gul_per_lb=self.ktools_num_gul_per_lb,
                        num_fm_per_lb=self.ktools_num_fm_per_lb,
                        run_debug=self.verbose,
                        stderr_guard=not self.ktools_disable_guard,
                        gul_legacy_stream=self.ktools_legacy_stream,
                        fifo_tmp_dir=not self.ktools_fifo_relative,
                        custom_gulcalc_cmd=self.model_custom_gulcalc,
                        custom_gulcalc_log_start=self.model_custom_gulcalc_log_start,
                        custom_gulcalc_log_finish=self.model_custom_gulcalc_log_finish,
                        gulpy=(self.gulpy and not self.model_custom_gulcalc),
                        gulpy_random_generator=self.gulpy_random_generator,
                        gulmc=(self.gulmc and not self.model_custom_gulcalc and not self.gulpy),
                        gulmc_random_generator=self.gulmc_random_generator,
                        gulmc_effective_damageability=self.gulmc_effective_damageability,
                        gulmc_vuln_cache_size=self.gulmc_vuln_cache_size,
                        fmpy=self.fmpy,
                        fmpy_low_memory=self.fmpy_low_memory,
                        fmpy_sort_output=self.fmpy_sort_output,
                        event_shuffle=self.ktools_event_shuffle,
                        modelpy=self.modelpy,
                        model_py_server=self.model_py_server,
                        peril_filter=self._get_peril_filter(analysis_settings),
                        summarypy=self.summarypy,
                        model_df_engine=self.model_df_engine or self.base_df_engine,
                        dynamic_footprint=self.dynamic_footprint
                    )
                except TypeError:
                    warnings.simplefilter("always")
                    warnings.warn(
                        f"{package_name}.supplier_model_runner doesn't accept new runner arguments, please add **kwargs to the run function signature")
                    model_runner_module.run(
                        analysis_settings,
                        number_of_processes=self.ktools_num_processes,
                        filename=script_fp,
                        num_reinsurance_iterations=ri_layers,
                        set_alloc_rule_gul=self.ktools_alloc_rule_gul,
                        set_alloc_rule_il=self.ktools_alloc_rule_il,
                        set_alloc_rule_ri=self.ktools_alloc_rule_ri,
                        run_debug=self.verbose,
                        stderr_guard=not self.ktools_disable_guard,
                        gul_legacy_stream=self.ktools_legacy_stream,
                        fifo_tmp_dir=not self.ktools_fifo_relative,
                        custom_gulcalc_cmd=self.model_custom_gulcalc
                    )

            except CalledProcessError as e:
                bash_trace_fp = os.path.join(model_run_fp, 'log', 'bash.log')
                if os.path.isfile(bash_trace_fp):
                    with io.open(bash_trace_fp, 'r', encoding='utf-8') as f:
                        self.logger.info('\nBASH_TRACE:\n' + "".join(f.readlines()))

                stderror_fp = os.path.join(model_run_fp, 'log', 'stderror.err')
                if os.path.isfile(stderror_fp):
                    with io.open(stderror_fp, 'r', encoding='utf-8') as f:
                        self.logger.info('\nKTOOLS_STDERR:\n' + "".join(f.readlines()))

                gul_stderror_fp = os.path.join(model_run_fp, 'log', 'gul_stderror.err')
                if os.path.isfile(gul_stderror_fp):
                    with io.open(gul_stderror_fp, 'r', encoding='utf-8') as f:
                        self.logger.info('\nGUL_STDERR:\n' + "".join(f.readlines()))

                self.logger.info('\nSTDOUT:\n' + e.output.decode('utf-8').strip())

                raise OasisException(
                    'Ktools run Error: non-zero exit code or error/warning messages detected in STDERR output.\n'
                    'Killing all processes. To disable this automated check run with `--ktools-disable-guard`.\n'
                    'Logs stored in: {}/log'.format(model_run_fp)
                )

        self.logger.info('Losses generated in {}'.format(model_run_fp))
        return model_run_fp


class GenerateLossesDeterministic(ComputationStep):
    step_params = [
        {'name': 'oasis_files_dir', 'is_path': True, 'pre_exist': True},
        {'name': 'output_dir', 'default': None},
        {'name': 'include_loss_factor', 'default': True},
        {'name': 'loss_factor', 'default': [1.0]},
        {'name': 'net_ri', 'default': False},
        {'name': 'ktools_alloc_rule_il', 'default': KTOOLS_ALLOC_IL_DEFAULT},
        {'name': 'ktools_alloc_rule_ri', 'default': KTOOLS_ALLOC_RI_DEFAULT},
        {'name': 'fmpy', 'default': True},
        {'name': 'fmpy_low_memory', 'default': False},
        {'name': 'fmpy_sort_output', 'default': False},
        {'name': 'il_stream_type', 'default': 2},
    ]

    def run(self):

        losses = OrderedDict({'gul': None, 'il': None, 'ri': None})
        output_dir = self.output_dir or self.oasis_files_dir

        il = all(p in os.listdir(self.oasis_files_dir) for p in ['fm_policytc.csv', 'fm_profile.csv', 'fm_programme.csv', 'fm_xref.csv'])
        ri = any(re.match(r'RI_\d+$', fn) for fn in os.listdir(self.oasis_files_dir))

        step_flag = ''
        try:
            pd.read_csv(os.path.join(self.oasis_files_dir, 'fm_profile.csv'))['step_id']
        except (OSError, FileNotFoundError, KeyError):
            pass
        else:
            step_flag = '-S'

        csv_to_bin(self.oasis_files_dir, output_dir, il=il, ri=ri)

        # Generate an items and coverages dataframe and set column types (important!!)
        items = merge_dataframes(
            pd.read_csv(os.path.join(self.oasis_files_dir, 'items.csv')),
            pd.read_csv(os.path.join(self.oasis_files_dir, 'coverages.csv')),
            on=['coverage_id'], how='left'
        )

        dtypes = {t: ('uint32' if t != 'tiv' else 'float32') for t in items.columns}
        items = set_dataframe_column_dtypes(items, dtypes)
        items['count'] = items['coverage_id'].map(items['coverage_id'].value_counts())

        items['tiv'] = items['tiv'] / items['count']
        items.drop(columns=['count'], inplace=True)
        # Change order of stream depending on rule type
        #   Stream_type 1
        #     event_id, item_id, sidx, loss
        #     1,1,-1,0
        #     1,1,-2,0
        #     1,1,-3,10000
        #     1,1,1,10000
        #
        #   Stream_type 2
        #     event_id, item_id, sidx, loss
        #     1,1,-3,10000
        #     1,1,-2,0
        #     1,1,-1,0
        #     1,1,1,10000
        if self.il_stream_type == 2:
            gulcalc_sidxs = \
                [KTOOLS_TIV_SAMPLE_IDX, KTOOLS_STD_DEV_SAMPLE_IDX, KTOOLS_MEAN_SAMPLE_IDX] + \
                list(range(1, len(self.loss_factor) + 1))
        elif self.il_stream_type == 1:
            gulcalc_sidxs = \
                [KTOOLS_MEAN_SAMPLE_IDX, KTOOLS_STD_DEV_SAMPLE_IDX, KTOOLS_TIV_SAMPLE_IDX] + \
                list(range(1, len(self.loss_factor) + 1))
        else:
            OasisException("Unknown il stream type: {}".format(self.il_stream_type))

        # Set damage percentages corresponing to the special indexes.
        # We don't care about mean and std_dev, but
        # TIV needs to be set correctly.
        special_loss_factors = {
            KTOOLS_MEAN_SAMPLE_IDX: 0.,
            KTOOLS_STD_DEV_SAMPLE_IDX: 0.,
            KTOOLS_TIV_SAMPLE_IDX: 1.
        }

        loss_factor_map = {**special_loss_factors, **{i + 1: val for i, val in enumerate(self.loss_factor)}}

        guls = items[['item_id', 'tiv']].join(pd.DataFrame({'sidx': gulcalc_sidxs}, dtype='int64'), how='cross').assign(event_id=1)
        guls['loss'] = guls['sidx'].map(loss_factor_map) * guls['tiv']
        guls = guls.astype({
            'event_id': int,
            'item_id': int,
            'sidx': int,
            'loss': float})[['event_id', 'item_id', 'sidx', 'loss']]
        guls_fp = os.path.join(output_dir, "raw_guls.csv")
        guls.to_csv(guls_fp, index=False)

        # il_stream_type = 2 if self.fmpy else 1
        ils_fp = os.path.join(output_dir, 'raw_ils.csv')

        # Create IL fmpy financial structures
        if self.fmpy:
            with setcwd(self.oasis_files_dir):
                check_call(f"{get_fmcmd(self.fmpy)} -a {self.ktools_alloc_rule_il} --create-financial-structure-files -p {output_dir}", shell=True)

        cmd = 'gultobin -S {} -t {} < {} | {} -p {} -a {} {} | tee ils.bin | fmtocsv > {}'.format(
            len(self.loss_factor),
            self.il_stream_type,
            guls_fp,
            get_fmcmd(self.fmpy, self.fmpy_low_memory, self.fmpy_sort_output),
            output_dir,
            self.ktools_alloc_rule_il,
            step_flag, ils_fp
        )

        try:
            self.logger.debug("RUN: " + cmd)
            check_call(cmd, shell=True)
        except CalledProcessError as e:
            raise OasisException("Exception raised in 'generate_deterministic_losses'", e)

        guls.drop(guls[guls['sidx'] < 1].index, inplace=True)
        guls.reset_index(drop=True, inplace=True)
        if self.include_loss_factor:
            guls['loss_factor_idx'] = guls.apply(
                lambda r: int(r['sidx'] - 1), axis='columns')
        guls.drop('sidx', axis=1, inplace=True)
        guls = guls[(guls[['loss']] != 0).any(axis=1)]

        losses['gul'] = guls

        ils = get_dataframe(src_fp=ils_fp, lowercase_cols=False)
        ils.drop(ils[ils['sidx'] < 0].index, inplace=True)
        ils.reset_index(drop=True, inplace=True)
        if self.include_loss_factor:
            ils['loss_factor_idx'] = ils.apply(
                lambda r: int(r['sidx'] - 1), axis='columns')
        ils.drop('sidx', axis=1, inplace=True)
        ils = ils[(ils[['loss']] != 0).any(axis=1)]
        losses['il'] = ils

        if ri:
            try:
                [fn for fn in os.listdir(self.oasis_files_dir) if fn == 'ri_layers.json'][0]
            except IndexError:
                raise OasisException(
                    'No RI layers JSON file "ri_layers.json " found in the '
                    'input directory despite presence of RI input files'
                )
            else:
                try:
                    with io.open(os.path.join(self.oasis_files_dir, 'ri_layers.json'), 'r', encoding='utf-8') as f:
                        ri_layers = len(json.load(f))
                except (IOError, JSONDecodeError, OSError, TypeError) as e:
                    raise OasisException('Error trying to read the RI layers file: {}'.format(e))
                else:
                    def run_ri_layer(layer):
                        layer_inputs_fp = os.path.join(output_dir, 'RI_{}'.format(layer))
                        # Create RI fmpy financial structures
                        if self.fmpy:
                            with setcwd(self.oasis_files_dir):
                                check_call(
                                    f"{get_fmcmd(self.fmpy)} -a {self.ktools_alloc_rule_ri} --create-financial-structure-files -p {layer_inputs_fp}",
                                    shell=True)

                        _input = 'gultobin -S 1 -t {} < {} | {} -p {} -a {} {} | tee ils.bin |'.format(
                            self.il_stream_type,
                            guls_fp,
                            get_fmcmd(self.fmpy, self.fmpy_low_memory, self.fmpy_sort_output),
                            output_dir,
                            self.ktools_alloc_rule_il,
                            step_flag
                        ) if layer == 1 else ''
                        pipe_in_previous_layer = '< ri{}.bin'.format(layer - 1) if layer > 1 else ''
                        ri_layer_fp = os.path.join(output_dir, 'ri{}.csv'.format(layer))
                        net_flag = "-n" if self.net_ri else ""
                        cmd = '{} {} -p {} {} -a {} {} {} | tee ri{}.bin | fmtocsv > {}'.format(
                            _input,
                            get_fmcmd(self.fmpy, self.fmpy_low_memory, self.fmpy_sort_output),
                            layer_inputs_fp,
                            net_flag,
                            self.ktools_alloc_rule_ri,
                            pipe_in_previous_layer,
                            step_flag,
                            layer,
                            ri_layer_fp
                        )
                        try:
                            self.logger.debug("RUN: " + cmd)
                            check_call(cmd, shell=True)
                        except CalledProcessError as e:
                            raise OasisException("Exception raised in 'generate_deterministic_losses'", e)
                        rils = get_dataframe(src_fp=ri_layer_fp, lowercase_cols=False)
                        rils.drop(rils[rils['sidx'] < 0].index, inplace=True)
                        if self.include_loss_factor:
                            rils['loss_factor_idx'] = rils.apply(
                                lambda r: int(r['sidx'] - 1), axis='columns')

                        rils.drop('sidx', axis=1, inplace=True)
                        rils.reset_index(drop=True, inplace=True)
                        rils = rils[(rils[['loss']] != 0).any(axis=1)]

                        return rils

                    for i in range(1, ri_layers + 1):
                        rils = run_ri_layer(i)
                        if i in [1, ri_layers]:
                            losses['ri'] = rils

        return losses


class GenerateLossesDummyModel(GenerateDummyOasisFiles):
    step_params = [
        {'name': 'analysis_settings_json', 'flag': '-z', 'is_path': True, 'pre_exist': True, 'required': True,
         'help': 'Analysis settings JSON file path'},
        {'name': 'ktools_num_processes', 'flag': '-n', 'type': int, 'default': KTOOLS_NUM_PROCESSES,
         'required': False, 'help': 'Number of ktools calculation processes to use'},
        {'name': 'ktools_alloc_rule_gul', 'type': int, 'default': KTOOLS_ALLOC_GUL_DEFAULT,
         'required': False, 'help': 'Set the allocation rule used in gulcalc'},
        {'name': 'ktools_alloc_rule_il', 'type': int, 'default': KTOOLS_ALLOC_IL_DEFAULT,
         'required': False, 'help': 'Set the fmcalc allocation rule used in direct insured loss'}
    ]
    chained_commands = [GenerateDummyModelFiles, GenerateDummyOasisFiles]

    def _validate_input_arguments(self):
        super()._validate_input_arguments()
        alloc_ranges = {
            'ktools_alloc_rule_gul': KTOOLS_ALLOC_GUL_MAX,
            'ktools_alloc_rule_il': KTOOLS_ALLOC_FM_MAX
        }
        for rule in alloc_ranges:
            alloc_val = getattr(self, rule)
            if alloc_val < 0 or alloc_val > alloc_ranges[rule]:
                raise OasisException(f'Error: {rule}={alloc_val} - Not within valid range [0..{alloc_ranges[rule]}]')

    def _validate_analysis_settings(self):
        warnings.simplefilter('always')

        # RI output is unsupported
        if self.analysis_settings.get('ri_output'):
            warnings.warn('Generating reinsurance losses with dummy model not currently supported. Ignoring ri_output in analysis settings JSON.')
            self.analysis_settings['ri_output'] = False

        # No loc or acc files so grouping losses based on OED fields is
        # unsupported
        # No generation of return periods file so currently unsupported
        loss_types = [False, False]
        loss_params = [
            {
                'loss': 'GUL', 'output': 'gul_output',
                'summary': 'gul_summaries', 'num_summaries': 0
            }, {
                'loss': 'IL', 'output': 'il_output',
                'summary': 'il_summaries', 'num_summaries': 0
            }
        ]
        for idx, param in enumerate(loss_params):
            if self.analysis_settings.get(param['output']):
                param['num_summaries'] = len(self.analysis_settings[param['summary']])
                self.analysis_settings[param['summary']][:] = [x for x in self.analysis_settings[param['summary']] if not x.get('oed_fields')]
                num_dropped_summaries = param['num_summaries'] - len(self.analysis_settings[param['summary']])
                if num_dropped_summaries == param['num_summaries']:
                    warnings.warn(
                        f'Grouping losses based on OED fields is unsupported. No valid {param["loss"]} output. Please change {param["loss"]} settings in analysis settings JSON.')
                    self.analysis_settings[param['output']] = False
                elif num_dropped_summaries > 0:
                    warnings.warn(
                        f'Grouping losses based on OED fields is unsupported. {num_dropped_summaries} groups ignored in {param["loss"]} output.')
                if param['num_summaries'] > 1:  # Get first summary only
                    self.analysis_settings[param['summary']] = [self.analysis_settings[param['summary']][0]]
                if self.analysis_settings[param['output']]:
                    # We should only have one summary now
                    self.analysis_settings[param['summary']][0]['id'] = 1
                    if self.analysis_settings[param['summary']][0]['leccalc']['return_period_file']:
                        warnings.warn('Return period file is not generated. Please use "return_periods" field in analysis settings JSON.')
                        self.analysis_settings[param['summary']][0]['leccalc']['return_period_file'] = False
                    loss_types[idx] = True
        (self.gul, self.il) = loss_types

        # Check for valid outputs
        if not any([self.gul, self.il]):
            raise OasisException('No valid output settings. Please check analysis settings JSON.')
        if not self.gul:
            raise OasisException('Valid GUL output required. Please check analysis settings JSON.')

        # Determine whether random number file will exist
        if self.analysis_settings.get('model_settings').get('use_random_number_file'):
            if self.num_randoms == 0:
                warnings.warn('Ignoring use random number file option in analysis settings JSON as no random number file will be generated.')
                self.analysis_settings['model_settings']['use_random_number_file'] = False

    def _prepare_run_directory(self):
        super()._prepare_run_directory()
        self.output_dir = os.path.join(self.target_dir, 'output')
        self.work_dir = os.path.join(self.target_dir, 'work')
        directories = [
            self.output_dir, self.work_dir
        ]
        for directory in directories:
            if not os.path.exists(directory):
                Path(directory).mkdir(parents=True, exist_ok=True)

        # Write new analysis_settings.json to target directory
        analysis_settings_fp = os.path.join(
            self.target_dir, 'analysis_settings.json'
        )
        with open(analysis_settings_fp, 'w') as f:
            json.dump(self.analysis_settings, f, indent=4, ensure_ascii=False)

    def _write_summary_info_files(self):
        summary_info_df = pd.DataFrame(
            {'summary_id': [1], '_not_set_': ['All-Risks']}
        )
        summary_info_fp = [
            os.path.join(self.output_dir, 'gul_S1_summary-info.csv'),
            os.path.join(self.output_dir, 'il_S1_summary-info.csv')
        ]
        for fp, loss_type in zip(summary_info_fp, [self.gul, self.il]):
            if loss_type:
                summary_info_df.to_csv(
                    path_or_buf=fp, encoding='utf-8', index=False
                )

    def run(self):
        self.logger.info('\nProcessing arguments - Creating Model & Test Oasis Files')

        self._validate_input_arguments()
        self.analysis_settings = AnalysisSettingSchema().get(
            self.analysis_settings_json
        )
        self._validate_analysis_settings()
        self._create_target_directory(label='losses')
        self._prepare_run_directory()
        self._get_model_file_objects()
        self._get_gul_file_objects()

        self.il = False
        if self.analysis_settings.get('il_output'):
            self.il = True
            self._get_fm_file_objects()
        else:
            self.fm_files = []

        output_files = self.model_files + self.gul_files + self.fm_files
        for output_file in output_files:
            output_file.write_file()

        self.logger.info(f'\nGenerating losses (GUL=True, IL={self.il})')

        self._write_summary_info_files()
        if self.ktools_num_processes == KTOOLS_NUM_PROCESSES:
            self.ktools_num_processes = multiprocessing.cpu_count()
        script_fp = os.path.join(self.target_dir, 'run_ktools.sh')
        bash.genbash(
            max_process_id=self.ktools_num_processes,
            analysis_settings=self.analysis_settings,
            gul_alloc_rule=self.ktools_alloc_rule_gul,
            il_alloc_rule=self.ktools_alloc_rule_il,
            filename=script_fp
        )
        bash_trace = subprocess.check_output(['bash', script_fp])
        self.logger.info(bash_trace.decode('utf-8'))

        self.logger.info(f'\nDummy Model run completed successfully in {self.target_dir}')
